"""
Base functionality for reading, writing, determining file formats, and scanning
Das Data.
"""
import os.path
from abc import ABC
from collections import defaultdict
from functools import cache, cached_property, reduce
from importlib.metadata import entry_points
from operator import add
from pathlib import Path
from typing import List, Optional, Union

import pandas as pd
from typing_extensions import Self

import dascore as dc
from dascore.constants import PatchType, SpoolType, timeable_types
from dascore.core.schema import PatchFileSummary
from dascore.exceptions import (
    DASCoreError,
    InvalidFiberFile,
    InvalidFileFormatter,
    UnknownFiberFormat,
)
from dascore.utils.docs import compose_docstring
from dascore.utils.hdf5 import HDF5ExtError
from dascore.utils.misc import suppress_warnings
from dascore.utils.patch import scan_patches
from dascore.utils.pd import list_ser_to_str


class _FiberIOManager:
    """
    A structure for intelligently storing, loading, and return FiberIO objects.

    This should only be used in conjunction with `FiberIO`.
    """

    def __init__(self, entry_point: str):
        self._entry_point = entry_point
        self._loaded_eps: set[str] = set()
        self._format_version = defaultdict(dict)
        self._extension_list = defaultdict(list)

    @cached_property
    def _eps(self):
        """
        Get the unlaoded entry points registered to this domain into a dict of
        {name: ep}
        """
        # TODO remove warning suppression and switch to select when 3.9 is dropped
        # see https://docs.python.org/3/library/importlib.metadata.html#entry-points
        with suppress_warnings(DeprecationWarning):
            out = {ep.name: ep.load for ep in entry_points()[self._entry_point]}
        return pd.Series(out)

    @cached_property
    def known_formats(self):
        """Return names of known formats."""
        formats = self._eps.index.str.split("__").str[0]
        return set(formats) | set(self._format_version)

    @property
    def unloaded_formats(self):
        """Return names of known formats."""
        return sorted(self.known_formats - set(self._format_version))

    @cached_property
    def _prioritized_list(self):
        """Yield a prioritized list of formatters."""
        # must load all plugins before getting list
        self.load_plugins()
        priority_formatters = []
        second_class_formatters = []
        for format_name in self.known_formats:
            unsorted = self._format_version[format_name]
            keys = sorted(unsorted, reverse=True)
            formatters = [unsorted[key] for key in keys]
            priority_formatters.append(formatters[0])
            if len(formatters) > 1:
                second_class_formatters.extend(formatters[1:])
        return tuple(priority_formatters + second_class_formatters)

    @cache
    def load_plugins(self, format: Optional[str] = None):
        """Load plugin for specific format or ensure all formats are loaded"""
        if format is not None and format in self._format_version:
            return  # already loaded
        if not (unloaded := self.unloaded_formats):
            return
        formats = {format} if format is not None else unloaded
        # load one, or all, formats
        for form in formats:
            for eps in self._eps.loc[self._eps.index.str.startswith(form)]:
                self.register_fiberio(eps()())
        # The selected format(s) should now be loaded
        assert set(formats).isdisjoint(self.unloaded_formats)

    def register_fiberio(self, fiberio: "FiberIO"):
        """Register a new fiber IO to manage."""
        forma, ver = fiberio.name.upper(), fiberio.version
        self._loaded_eps.add(fiberio.name)
        for ext in iter(fiberio.preferred_extensions):
            self._extension_list[ext].append(fiberio)
        self._format_version[forma][ver] = fiberio

    def get_fiberio(
        self,
        format: Optional[str] = None,
        version: Optional[str] = None,
        extension: Optional[str] = None,
    ) -> "FiberIO":
        """
        Return the most likely formatter for given inputs.

        If no such formatter exists, raise UnknownFiberFormat error.

        Parameters
        ----------
        format
            The format string indicating the format name
        version
            The version string of the format
        extension
            The extension of the file.
        """
        iterator = self.yield_fiberio(
            format=format,
            version=version,
            extension=extension,
        )
        for formatter in iterator:
            return formatter

    def yield_fiberio(
        self,
        format: Optional[str] = None,
        version: Optional[str] = None,
        extension: Optional[str] = None,
    ) -> Self:
        """
        Yields fiber IO object based on input priorities.

        The order is sorted in likelihood of the formatter being correct. For
        example, if file format is specified but file_version is not, all
        formatters for the format will be yielded with the newest versions
        first in the list.

        If neither version nor format are specified but extension is all formatters
        specifying the extension will be first in the list, sorted by format name
        and format version.

        If nothing is specified, all formatters will be returned starting with
        the newest (the highest version) of each formatter, followed by older
        versions.

        Parameters
        ----------
        format
            The format string indicating the format name
        version
            The version string of the format
        extension
            The extension of the file.
        """
        # TODO replace this with concise pattern matching once 3.9 is dropped
        if version and not format:
            msg = "Providing only a version is not sufficient to determine format"
            raise UnknownFiberFormat(msg)
        if format is not None:
            self.load_plugins(format)
            yield from self._yield_format_version(format, version)
        elif extension is not None:
            yield from self._yield_extensions(extension)
        else:
            yield from self._prioritized_list

    def _yield_format_version(self, format, version):
        """Yield file format/version prioritized formatters."""
        if format is not None:
            format = format.upper()
            self.load_plugins(format)
            formatters = self._format_version.get(format, None)
            # no format found
            if not formatters:
                format_list = list(self.known_formats)
                msg = f"Unknown format {format}, " f"known formats are {format_list}"
                raise UnknownFiberFormat(msg)
            # a version is specified
            if version:
                formatter = formatters.get(version, None)
                if formatter is None:
                    msg = (
                        f"Format {format} has no version: [{version}] "
                        f"known versions of this format are: {list(formatters)}"
                    )
                    raise UnknownFiberFormat(msg)
                yield formatter
                return
            # reverse sort formatters and yield latest version first.
            for formatter in dict(sorted(formatters.items(), reverse=True)).values():
                yield formatter
            return

    def _yield_extensions(self, extension):
        """generator to get formatter prioritized by preferred extensions."""
        has_yielded = set()
        self.load_plugins()
        for formatter in self._extension_list[extension]:
            yield formatter
            has_yielded.add(formatter)
        for formatter in self._prioritized_list:
            if formatter not in has_yielded:
                yield formatter


# ------------- Protocol for File Format support


class FiberIO(ABC):
    """
    An interface which adds support for a given filer format.

    This class should be subclassed when adding support for new formats.
    """

    name: str = ""
    version: str = ""
    preferred_extensions: tuple[str] = ()
    manager = _FiberIOManager("dascore.fiber_io")

    def read(self, path, **kwargs) -> SpoolType:
        """
        Load data from a path.

        *kwargs should include support for selecting expected dimensions. For
        example, distance=(100, 200) would only read data with distance from
        100 to 200.
        """
        msg = f"FileFormatter: {self.name} has no read method"
        raise NotImplementedError(msg)

    def scan(self, path) -> List[PatchFileSummary]:
        """
        Returns a list of summary info for patches contained in file.
        """
        # default scan method reads in the file and returns required attributes
        # however, this can be very slow, so each parser should implement scan
        # when possible.
        try:
            spool = self.read(path)
        except NotImplementedError:
            msg = f"FileFormatter: {self.name} has no scan or read method"
            raise NotImplementedError(msg)
        out = []
        for pa in spool:
            info = dict(pa.attrs)
            info["file_format"] = self.name
            info["path"] = str(path)
            out.append(PatchFileSummary.parse_obj(info))
        return out

    def write(self, spool: SpoolType, path: Union[str, Path]):
        """
        Write the spool to disk
        """
        msg = f"FileFormatter: {self.name} has no write method"
        raise NotImplementedError(msg)

    def get_format(self, path) -> Union[tuple[str, str], bool]:
        """
        Return a tuple of (format_name, version_numbers).

        This should only work if path is the supported file format, otherwise
        raise UnknownFiberError or return False.
        """
        msg = f"FileFormatter: {self.name} has no get_version method"
        raise NotImplementedError(msg)

    @property
    def implements_scan(self) -> bool:
        """
        Returns True if the subclass implements its own scan method else False.
        """
        return self.scan.__func__ is not FiberIO.scan

    @property
    def implements_get_format(self) -> bool:
        """Return True if the subclass implements its own get_format method."""
        return self.get_format.__func__ is not FiberIO.get_format

    def __hash__(self):
        """FiberIO instances should be uniquely defined by (format, version)"""
        return hash((self.name, self.version))

    def __init_subclass__(cls, **kwargs):
        """
        Hook for registering subclasses.
        """
        # check that the subclass is valid
        if not cls.name:
            msg = "You must specify the file format with the name field."
            raise InvalidFileFormatter(msg)
        # register formatter
        manager: _FiberIOManager = getattr(cls.__mro__[1], "manager")
        manager.register_fiberio(cls())


def read(
    path: Union[str, Path],
    file_format: Optional[str] = None,
    file_version: Optional[str] = None,
    time: Optional[tuple[Optional[timeable_types], Optional[timeable_types]]] = None,
    distance: Optional[tuple[Optional[float], Optional[float]]] = None,
    **kwargs,
) -> SpoolType:
    """
    Read a fiber file.

    Parameters
    ----------
    path
        A path to the file to read.
    file_format
        A string indicating the file format. If not provided dascore will
        try to estimate the format.
    file_version
        An optional string indicating the format version.
    time
        An optional tuple of time ranges.
    distance
        An optional tuple of distances.
    *kwargs
        All kwargs are passed to the format-specific read functions.
    """
    if not file_format or not file_version:
        file_format, file_version = get_format(
            path,
            file_format=file_format,
            file_version=file_version,
        )
    formatter = FiberIO.manager.get_fiberio(file_format, file_version)
    return formatter.read(
        path, file_version=file_version, time=time, distance=distance, **kwargs
    )


@compose_docstring(fields=list(PatchFileSummary.__fields__))
def scan_to_df(
    path: Union[Path, str, PatchType, SpoolType],
    file_format: Optional[str] = None,
    file_version: Optional[str] = None,
    ignore: bool = False,
) -> pd.DataFrame:
    """
    Scan a path, return a dataframe of contents.

    Parameters
    ----------
    path
        The path the to file to scan
    file_format
        Format of the file. If not provided DASCore will try to determine it.
    ignore
        If True, ignore non-DAS files by returning an empty list, else raise
        UnknownFiberFormat if unreadable file encountered.

    Returns
    -------
    Return a dataframe with columns:
        {fields}
    """
    info = scan(
        path_or_spool=path,
        file_format=file_format,
        file_version=file_version,
        ignore=ignore,
    )
    df = pd.DataFrame([dict(x) for x in info]).assign(
        dims=lambda x: list_ser_to_str(x["dims"])
    )
    return df[list(PatchFileSummary.get_index_columns())]


@compose_docstring(fields=list(PatchFileSummary.__annotations__))
def scan(
    path_or_spool: Union[Path, str, PatchType, SpoolType],
    file_format: Optional[str] = None,
    file_version: Optional[str] = None,
    ignore: bool = False,
) -> list[PatchFileSummary]:
    """
    Scan a file, return the summary dictionary.

    Parameters
    ----------
    path_or_spool
        The path the to file to scan
    file_format
        Format of the file. If not provided DASCore will try to determine it.
        Only applicable for path-like inputs.
    file_version
        Version of the file. If not provided DASCore will try to determine it.
        Only applicable for path-like inputs.
    ignore
        If True, ignore non-DAS files by returning an empty list, else raise
        UnknownFiberFormat if unreadable file encountered.

    Notes
    -----
    The summary dictionaries contain the following fields:
        {fields}
    """
    if isinstance(path_or_spool, (str, Path)):
        out = _scan_from_path(
            path_or_spool,
            file_format=file_format,
            file_version=file_version,
            ignore=ignore,
        )
    else:
        out = scan_patches(path_or_spool)
    return out


def _scan_from_path(
    path: Union[Path, str, PatchType, SpoolType],
    file_format: Optional[str] = None,
    file_version: Optional[str] = None,
    ignore: bool = False,
) -> list[PatchFileSummary]:
    """Scan from a single path."""
    if not os.path.exists(path):
        msg = f"{path} does not exist"
        raise InvalidFiberFile(msg)
    # recursively handle directory files
    if os.path.isdir(path):
        out = [
            _scan_from_path(x, file_format, file_version, ignore=True)
            for x in Path(path).glob("*")
        ]
        if len(out):
            return reduce(add, out)
        else:
            return []
    # dispatch to file format handlers
    if not file_format or not file_version:
        try:
            file_format, file_version = get_format(
                path,
                file_format=file_format,
                file_version=file_version,
            )
        except UnknownFiberFormat as e:
            if ignore:
                return []
            else:
                raise e

    formatter = FiberIO.manager.get_fiberio(file_format, file_version)
    out = formatter.scan(path)
    return out


def get_format(
    path: Union[str, Path],
    file_format: Optional[str] = None,
    file_version: Optional[str] = None,
) -> tuple[str, str]:
    """
    Return the name of the format contained in the file and version number.

    Parameters
    ----------
    path
        The path to the file.
    file_format
        The known file format.
    file_version
        The known file version.


    Returns
    -------
    A tuple of (file_format_name, version) both as strings.

    Raises
    ------
    dascore.exceptions.UnknownFiberFormat - Could not determine the fiber format.

    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"{path} does not exist.")
    ext = Path(path).suffix or None
    iterator = FiberIO.manager.yield_fiberio(file_format, file_version, extension=ext)
    for formatter in iterator:
        try:
            format_version = formatter.get_format(path)
        except (ValueError, TypeError, HDF5ExtError, NotImplementedError, DASCoreError):
            continue
        if format_version:
            return format_version
    else:
        msg = f"Could not determine file format of {path}"
        raise UnknownFiberFormat(msg)


def write(
    patch_or_spool,
    path: Union[str, Path],
    file_format: str,
    file_version: Optional[str] = None,
    **kwargs,
) -> Path:
    """
    Write a Patch or Spool to disk.

    Parameters
    ----------
    path
        The path to the file.
    file_format
        The string indicating the format to write.
    file_version
        Optionally specify the version of the file, else use the latest
        version.

    Raises
    ------
    dascore.exceptions.UnknownFiberFormat - Could not determine the fiber format.
    """
    formatter = FiberIO.manager.get_fiberio(file_format, file_version)
    if not isinstance(patch_or_spool, dc.BaseSpool):
        patch_or_spool = dc.spool([patch_or_spool])
    formatter.write(patch_or_spool, path, **kwargs)
    return path



"""Tests for pandas utility functions."""
import numpy as np
import pandas as pd
import pydantic
import pytest

from dascore.utils.pd import adjust_segments, fill_defaults_from_pydantic, filter_df
from dascore.utils.time import to_datetime64, to_timedelta64


@pytest.fixture
def example_df_2():
    """create a simple df for testing. Example from Chris Albon."""
    time = to_datetime64("2020-01-03")
    time_min = [time + x * np.timedelta64(1, "s") for x in range(5)]
    time_max = time_min + np.timedelta64(10, "m")
    raw_data = {
        "first_name": ["Jason", "Molly", "Tina", "Jake", "Amy"],
        "last_name": ["Miller", "Jacobson", "Ali", "Milner", "Cooze"],
        "bp_min": [100, 120, 110, 85, 125],
        "bp_max": [110, 145, 121, 99, 165],
        "time_min": time_min,
        "time_max": time_max,
        "d_time": [np.timedelta64(1, "s") for _ in range(5)],
    }
    out = pd.DataFrame(raw_data, columns=list(raw_data))
    return out


class TestFilterDfBasic:
    """Tests for filtering dataframes."""

    @pytest.fixture
    def example_df(self):
        """create a simple df for testing. Example from Chris Albon."""
        raw_data = {
            "first_name": ["Jason", "Molly", "Tina", "Jake", "Amy"],
            "last_name": ["Miller", "Jacobson", "Ali", "Milner", "Cooze"],
            "age": [42, 52, 36, 24, 73],
            "pre_test_score": [4, 24, 31, 2, 3],
            "post_test_score": [25, 94, 57, 62, 70],
        }
        return pd.DataFrame(raw_data, columns=list(raw_data))

    def test_string_basic(self, example_df):
        """test that specifying a string with no matching works."""
        out = filter_df(example_df, first_name="Jason")
        assert out[0]
        assert not out[1:].any()

    def test_string_matching(self, example_df):
        """unix style matching should also work."""
        # test *
        out = filter_df(example_df, first_name="J*")
        assert {"Jason", "Jake"} == set(example_df[out].first_name)
        # test ???
        out = filter_df(example_df, first_name="J???")
        assert {"Jake"} == set(example_df[out].first_name)

    def test_str_sequence(self, example_df):
        """Test str sequences find values in sequence."""
        out = filter_df(example_df, last_name={"Miller", "Jacobson"})
        assert out[:2].all()
        assert not out[2:].any()

    def test_non_str_single_arg(self, example_df):
        """test that filter index can be used on Non-nslc columns."""
        # test non strings
        out = filter_df(example_df, age=42)
        assert out[0]
        assert not out[1:].any()

    def test_non_str_sequence(self, example_df):
        """ensure sequences still work for isin style comparisons."""
        out = filter_df(example_df, age={42, 52})
        assert out[:2].all()
        assert not out[2:].any()

    def test_bad_parameter_raises(self, example_df):
        """ensure passing a parameter that doesn't have a column raises."""
        with pytest.raises(KeyError, match="not found in df"):
            filter_df(example_df, bad_column=2)

    def test_bad_parameter_doesnt_raise(self, example_df):
        """A bad parameter shouldn't raise if told filter is told to ignore."""
        out = filter_df(example_df, bad_column=2, ignore_bad_kwargs=True)
        # Each row should be kept.
        assert np.all(out)

    def test_min(self, example_df):
        """Tests for using min parameter."""
        df = example_df[filter_df(example_df, age_max=40)]
        assert df.equals(example_df[example_df["age"] < 40])

    def test_max(self, example_df):
        """Tests for using max parameter."""
        df = example_df[filter_df(example_df, post_test_score_min=60)]
        df2 = example_df[example_df["post_test_score"] >= 60]
        assert df.equals(df2)

    def test_min_and_max(self, example_df):
        """Ensure min/max can be used together."""
        ser = example_df["age"]
        con = (ser >= 20) & (ser <= 30)
        out = filter_df(example_df, age_min=20, age_max=30)
        assert all(con == out)


class TestFilterDfAdvanced:
    """Tests for advanced filtering of dataframes."""

    def test_col(self, example_df_2):
        """Test column names that end with max"""
        vals = [100, 125]
        out = filter_df(example_df_2, bp_min=vals)
        assert np.all(out == example_df_2["bp_min"].isin(vals))

    def test_min_max_range_based_on_column(self, example_df_2):
        """
        Using just bp, we should be able to specify range which spans
        two columns.
        """
        val = (100, 120)
        out = filter_df(example_df_2, bp=val)
        max_too_small = example_df_2["bp_max"] < val[0]
        min_too_big = example_df_2["bp_min"] > val[1]
        in_range = ~(max_too_small | min_too_big)
        assert all(out == in_range)

    def test_time_query_all_open(self, example_df_2):
        """Test for open time interval"""
        out = filter_df(example_df_2, time=(None, None))
        assert out.all()

    def test_time_query_one_open(self, example_df_2):
        """Test for open time interval"""
        tmax = to_datetime64(example_df_2["time_max"].max() - np.timedelta64(1, "ns"))
        out = filter_df(example_df_2, time=(tmax, None))
        # just the last row should have been selected
        assert out.iloc[-1] and out.astype(int).sum() == 1

    def test_time_query_with_string(self, example_df_2):
        """Test for time query with a string."""
        tmax = to_datetime64(example_df_2["time_max"].max() - np.timedelta64(1, "ns"))
        out1 = filter_df(example_df_2, time=(str(tmax), None))
        # just the last row should have been selected
        assert out1.iloc[-1] and out1.astype(int).sum() == 1
        # also ensure the string can be second element, all rows should be selected
        out2 = filter_df(example_df_2, time=(None, str(tmax)))
        assert len(out2) == len(example_df_2)


class TestAdjustSegments:
    """Tests for adjusting segments of dataframes."""

    @pytest.fixture()
    def adjacent_df(self):
        """Create a ddtaframe with adjacent times."""
        time_mins = [
            np.datetime64("2020-01-01"),
            np.datetime64("2020-01-01T00:00:10.01"),
        ]
        time_maxs = [
            np.datetime64("2020-01-01T00:00:10.01"),
            np.datetime64("2020-01-01T00:00:20"),
        ]
        dtime = to_timedelta64(0.01)
        df = (
            pd.DataFrame(index=range(2))
            .assign(time_min=time_mins, time_max=time_maxs, d_time=dtime)
            .assign(distance_min=0, distance_max=100, d_distance=1, station="BOB")
        )
        return df

    def test_limits_changed(self, example_df_2):
        """Ensure limits of segments are changed when intersected by query."""
        df = example_df_2
        new_min = df["time_min"].min() + np.timedelta64(1, "s")
        new_max = df["time_max"].max() - np.timedelta64(1, "s")
        new = adjust_segments(df, time=(new_min, new_max))
        assert np.all(new["time_min"] >= new_min)
        assert np.all(new["time_max"] <= new_max)

    def test_multiple_kwargs(self, adjacent_df):
        """Ensure multiple dimensions can be adjusted with one function call."""
        time = [
            adjacent_df["time_max"].min() - to_timedelta64(5),
            adjacent_df["time_max"].max() - to_timedelta64(5),
        ]
        distance = (30, 50)
        out = adjust_segments(adjacent_df, time=time, distance=distance)
        assert len(adjacent_df) == len(out)
        assert (out["time_min"] >= time[0]).all()
        assert (out["time_max"] <= time[1]).all()
        assert (out["distance_min"] >= distance[0]).all()
        assert (out["distance_max"] <= distance[1]).all()

    def test_missing_interval_col_raises_keyerro(self, adjacent_df):
        """Ensure if an interval column is missing a KeyError is raised."""
        df = adjacent_df.drop(columns=["distance_min"])
        with pytest.raises(KeyError):
            _ = adjust_segments(df, distance=(100, 200))

    class TestFillDefaultsFromPydantic:
        """Tests for initing empty columns from pydanitc models."""

        class Model(pydantic.BaseModel):
            """Example basemodel"""

            int_with_default: int = 10
            str_with_default: str = "bob"
            float_with_default: float = 10.0
            float_no_default: float

        @pytest.fixture(scope="class")
        def simple_df(self):
            """Create a simple df for testing."""
            df = pd.DataFrame(index=range(10)).assign(
                int_with_default=20,
                str_with_default="bill",
                float_with_default=22.2,
                float_no_default=42.0,
            )
            return df

        def test_no_missing_does_nothing(self, simple_df):
            """Ensure the default gets filled."""
            out = fill_defaults_from_pydantic(simple_df, self.Model)
            assert out.equals(simple_df)

        def test_missing_with_default_fills_default(self, simple_df):
            """Ensure the default values are filled in when missing"""
            df = simple_df.drop(columns="int_with_default")
            out = fill_defaults_from_pydantic(df, self.Model)
            assert "int_with_default" in out.columns
            assert (out["int_with_default"] == 10).all()

        def test_missing_with_no_default_raises(self, simple_df):
            """Ensure Value error is raised if no default is there."""
            df = simple_df.drop(columns="float_no_default")
            with pytest.raises(ValueError, match="required value"):
                fill_defaults_from_pydantic(df, self.Model)



"""
Utilities for chunking dataframes.
"""
from typing import Collection, Optional, Union

import numpy as np
import pandas as pd

from dascore.constants import numeric_types, timeable_types
from dascore.exceptions import ParameterError
from dascore.utils.pd import (
    get_column_names_from_dim,
    get_dim_names_from_columns,
    get_interval_columns,
)
from dascore.utils.time import is_datetime64, to_datetime64, to_timedelta64


def get_intervals(
    start,
    stop,
    length,
    overlap=None,
    step=None,
    keep_partials=False,
):
    """
    Create a range of values with optional overlaps.

    Parameters
    ----------
    start
        The start of the interval.
    stop
        The end of the interval.
    length
        The length of the segments.
    overlap
        The overlap of the start of each interval with the end
        of the previous interval.
    step
        If not None, subtract step (the sampling interval) from the end
        values so that the intervals do not overlap by one sample.
    keep_partials
        If True, keep the segments which are smaller than chunksize.

    Returns
    -------
    A 2D array where first column is start and second column is end.
    """
    # when length is null just use entire length
    if pd.isnull(length):
        out = np.array([start, stop])
        if is_datetime64(start):
            out = to_datetime64(out)
        return np.atleast_2d(out)

    if is_datetime64(start):
        length = to_timedelta64(length)
    # get variable and perform checks
    overlap = length * 0 if not overlap else overlap
    step = length * 0 if not step else step
    assert overlap < length, "Overlap must be less than step"
    assert (stop - start) > length, "Range must be greater than step"
    # reference with no overlap
    new_step = length - overlap
    reference = np.arange(start, stop + new_step, step=new_step)
    ends = reference[:-1] + length - step
    starts = reference[:-1]
    # trim end to not surpass stop
    if ends[-1] > stop:
        if not keep_partials:
            ends, starts = ends[:-1], starts[:-1]
        else:
            ends[-1] = stop
    return np.stack([starts, ends]).T


class ChunkManager:
    """
    A class for managing the chunking of data defined in a dataframe.

    The chunk manager handles both splitting and joining of contiguous,
    or near-contiguous, blocks of data.

    Parameters
    ----------
    overlap
        The amount of overlap between each segment, starting with the end of
        first row. Negative values can be used for inducing gaps.
    group_columns
        A sequence of column names which should be used for sorting groups.
    keep_partial
        If True, keep segments which are shorter than chunk size (at end of
        contiguous blocks)
    tolerance
        The upper limit of a gap to tolerate in terms of the sampling
        along the desired dimension. E.G., the default value means entities
        with gaps <= 1.5 * d_{name} will be merged.
    **kawrgs
        kwargs specify the column along which to chunk. The key specifies the
        column along which to chunk, typically, `time` or `distance`, and the
        value specifies the chunk size. A value of None means to chunk on all
        available data (e.g. merge all data).

    Notes
    -----
    This class is used internally by `dc.Spool.chunk`.
    """

    def __init__(
        self,
        overlap: Optional[Union[timeable_types, numeric_types]] = None,
        group_columns: Optional[Collection[str]] = None,
        keep_partial=False,
        tolerance=1.5,
        **kwargs,
    ):
        self._overlap = overlap
        self._group_columns = group_columns
        self._keep_partials = keep_partial
        self._tolerance = tolerance
        self._name, self._value = self._validate_kwargs(kwargs)
        self._validate_chunker()

    def _validate_kwargs(self, kwargs):
        """Ensure kwargs is len one and has a valid"""
        assert len(kwargs) == 1
        ((key, value),) = kwargs.items()
        return key, value

    def _validate_chunker(self):
        """Ensure selected parameters are compatible."""
        # chunker is used for merging
        if pd.isnull(self._value):
            if self._keep_partials or self._overlap:
                msg = (
                    "When chunk value is None (ie Chunker is used for merging) "
                    "both _keep_partials and self._overlap must not be selected."
                )
                raise ParameterError(msg)

    def _get_continuity_group_number(self, start, stop, step) -> pd.Series:
        """Return a series of ints indicating continuity group."""
        # start by sorting according to start time
        arg_ser = start.argsort()
        args = arg_ser.index[arg_ser.values]
        start_sorted, stop_sorted, step_sorted = start[args], stop[args], step[args]
        # next get cummax of endtimes and detect gaps
        stop_cum_max = stop_sorted.cummax()
        end_markers = stop_cum_max.shift() + step_sorted * self._tolerance
        has_gap = start_sorted > end_markers
        group_num = has_gap.astype(int).cumsum()
        return group_num[start.index]

    def _get_duration_overlap(self, duration, start, step, overlap=None):
        """
        Get duration and overlap from kwargs.
        """
        overlap = overlap if overlap is not None else self._overlap
        # cast step to time delta if start is datetime
        if is_datetime64(start):
            step = to_timedelta64(step)
            overlap = to_timedelta64(overlap)
        over = overlap if not pd.isnull(overlap) else (step * 0).iloc[0]
        return duration, over

    def _create_df(self, df, name, start_stop, gnum):
        """Reconstruct the dataframe."""
        out = pd.DataFrame(start_stop, columns=[f"{name}_min", f"{name}_max"])
        merger = df.drop(columns=out.columns)
        for col in merger:
            vals = merger[col].unique()
            assert len(vals) == 1, "Haven't yet implemented non-homogenous merging"
            out[col] = vals[0]
        # add the group number for getting instruction df later
        out["_group"] = gnum
        return out

    def get_instruction_df(self, source_df, chunked_df):
        """
        Get a dataframe connecting the chunked dataframe to its origin.

        This is used to connect source data to desired data after chunking
        operation.

        Parameters
        ----------
        source_df
            The dataframe before chunking
        chunked_df
            The chunked dataframe (output of `chunk` method)
        """
        # the group column should exist and the chunked groups should be subset
        # of the source groups
        assert "_group" in source_df.columns and "_group" in chunked_df.columns
        chunked_groups = set(chunked_df["_group"])
        assert chunked_groups.issubset(set(source_df["_group"]))
        min_name, max_name = f"{self._name}_min", f"{self._name}_max"
        # iterate each group and create instruction df
        out = []
        for group, sub_chunk in chunked_df.groupby("_group"):
            sub_source = source_df[source_df["_group"] == group]
            c_start, c_stop, c_step = get_interval_columns(sub_source, self._name)
            new_start, new_stop, new_step = get_interval_columns(sub_chunk, self._name)
            dims = get_dim_names_from_columns(sub_source)
            cols2keep = get_column_names_from_dim(dims)
            # TODO need to think more about this, a naive implementation for now
            for start, stop, ind in zip(
                new_start.values, new_stop.values, new_stop.index
            ):
                too_late = c_start > stop
                too_early = c_stop < start
                in_range = ~(too_early | too_late)
                assert in_range.sum() > 0, "no original data source found!"
                sub_df = sub_source[in_range][cols2keep]
                # trim intervals to existing values
                sub_df.loc[sub_df[min_name] < start, min_name] = start
                sub_df.loc[sub_df[max_name] > stop, max_name] = stop
                sub_df.loc[:, "source_index"] = sub_df.index.values
                sub_df.loc[:, "current_index"] = ind
                out.append(sub_df)
        df = pd.concat(out, axis=0).reset_index(drop=True).set_index("source_index")
        return df

    def _get_group(self, df, start, stop, step):
        """
        Get the group designation for df. This accounts for both time intervals
        being consistent and group columns matching.
        """
        # TODO: Test if different stations bridge time span
        group_cont = self._get_continuity_group_number(start, stop, step)
        cols = [f"d_{self._name}"] + list(self._group_columns or [])
        columns = [x for x in cols if x in df.columns]
        col_groups = df.groupby(columns).ngroup()
        group = group_cont.astype(str) + "_" + col_groups.astype(str)
        return group

    def chunk(
        self,
        df: pd.DataFrame,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Chunk a dataframe into new contiguous segments.

        The dataframe must have column names {key}_max, {key}_min, and d_{key}
        where {key} is the key used in the kwargs.

        Parameters
        ----------
        df
            Input dataframe to chunk.

        Returns
        -------
        A tuple of the original dataframe with added column '_group' and an
        output dataframe with column '_group'. The _group column is used
        to link the two dataframes together.
        """
        if df.empty:  # empy df, do nothing
            return df.assign(_group=None), df.assign(_group=None)
        # get series of start/stop along requested dimension
        start, stop, step = get_interval_columns(df, self._name)
        dur, overlap = self._get_duration_overlap(self._value, start, step)
        # get group numbers
        group = self._get_group(df, start, stop, step)
        # get max, min for each group and expand
        group_mins = start.groupby(group).min()
        group_maxs = stop.groupby(group).max()
        # split/group dataframe into new chunks by iterating over each group.
        out = []
        for gnum in group.unique():
            start, stop = group_mins[gnum], group_maxs[gnum]
            current_df = df.loc[group[group == gnum].index]
            # reconstruct DF
            new_start_stop = get_intervals(
                start,
                stop,
                dur,
                overlap=overlap,
                step=step.iloc[0],
                keep_partials=self._keep_partials,
            )
            # create the newly chunked dataframe
            sub_new_df = self._create_df(current_df, self._name, new_start_stop, gnum)
            out.append(sub_new_df)

        out = pd.concat(out, axis=0).reset_index(drop=True)
        return df.assign(_group=group), out



"""
A module for applying transformation to Patches.

Transforms are defined as
"""

from dascore.utils.misc import MethodNameSpace

from .fft import rfft
from .spectro import spectrogram
from .strain import velocity_to_strain_rate


class TransformPatchNameSpace(MethodNameSpace):
    """Patch namesapce for transformations."""

    velocity_to_strain_rate = velocity_to_strain_rate
    spectrogram = spectrogram
    rfft = rfft



"""
dascore Filtering module.

Much of this code was inspired by ObsPy's filtering module created by:
Tobias Megies, Moritz Beyreuther, Yannik Behr
"""

from typing import Sequence

import numpy as np
import pandas as pd
from scipy import ndimage
from scipy.signal import iirfilter, medfilt2d, sosfilt, sosfiltfilt, zpk2sos

import dascore
from dascore.constants import PatchType
from dascore.exceptions import FilterValueError
from dascore.utils.patch import patch_function


def _check_filter_kwargs(kwargs):
    """Check filter kwargs and return dim name and filter range."""
    if len(kwargs) != 1:
        msg = "pass filter requires you specify one dimension and filter range."
        raise FilterValueError(msg)
    dim = list(kwargs.keys())[0]
    filt_range = kwargs[dim]
    if not isinstance(filt_range, Sequence) or len(filt_range) != 2:
        msg = f"filter range must be a length two sequence not {filt_range}"
        raise FilterValueError(msg)
    filt1, filt2 = filt_range
    if all([pd.isnull(x) for x in filt_range]):
        msg = (
            f"pass filter requires at least one filter limit, "
            f"you passed {filt_range}"
        )
        raise FilterValueError(msg)

    return dim, filt1, filt2


def _check_sobel_args(dim, mode, cval):
    """Check Sobel filter kwargs and return"""
    mode_options = {
        "reflect",
        "constant",
        "nearest",
        "mirror",
        "wrap",
        "grid-constant",
        "grid-mirror",
        "grid-wrap",
    }
    if not isinstance(dim, str):
        msg = "dim parameter should be a string."
        raise FilterValueError(msg)
    if not isinstance(mode, str):
        msg = "mode parameter should be a string."
        raise FilterValueError(msg)
    if not isinstance(cval, (float, int)):
        msg = "cval parameter should be a float or an int."
        raise FilterValueError(msg)
    if mode not in mode_options:
        msg = f"The valid values for modes are {mode_options}."
        raise FilterValueError(msg)

    return dim, mode, cval


def _get_sampling_rate(patch, dim):
    """Get sampling rate, as a float from sampling period along a dimension."""
    d_dim = patch.attrs[f"d_{dim}"]
    if dim == "time":  # get time in to seconds
        d_dim = d_dim / np.timedelta64(1, "s")
    return 1.0 / d_dim


def _check_filter_range(niquest, low, high, filt_min, filt_max):
    """Simple check on filter parameters."""
    # ensure filter bounds are within niquest
    if low is not None and ((0 > low) or (low > 1)):
        msg = f"possible filter bounds are [0, {niquest}] you passed {filt_min}"
        raise FilterValueError(msg)
    if high is not None and ((0 > high) or (high > 1)):
        msg = f"possible filter bounds are [0, {niquest}] you passed {filt_max}"
        raise FilterValueError(msg)
    if high is not None and low is not None and high <= low:
        msg = (
            "Low filter param must be less than high filter param, you passed:"
            f"filt_min = {filt_min}, filt_max = {filt_max}"
        )
        raise FilterValueError(msg)


def _get_sos(sr, filt_min, filt_max, corners):
    """Get second order sections from sampling rate and filter bounds."""
    niquest = 0.5 * sr
    low = None if pd.isnull(filt_min) else filt_min / niquest
    high = None if pd.isnull(filt_max) else filt_max / niquest
    _check_filter_range(niquest, low, high, filt_min, filt_min)

    if (low is not None) and (high is not None):  # apply bandpass
        z, p, k = iirfilter(
            corners, [low, high], btype="band", ftype="butter", output="zpk"
        )
    elif low is not None:
        z, p, k = iirfilter(
            corners, low, btype="highpass", ftype="butter", output="zpk"
        )
    else:
        assert high is not None
        z, p, k = iirfilter(
            corners, high, btype="lowpass", ftype="butter", output="zpk"
        )
    return zpk2sos(z, p, k)


@patch_function()
def pass_filter(patch: PatchType, corners=4, zerophase=True, **kwargs) -> PatchType:
    """
    Apply a Butterworth pass filter (bandpass, highpass, or lowpass).

    Parameters
    ----------
    corners
        The number of corners for the filter.
    zerophase
        If True, apply the filter twice.
    **kwargs
        Used to specify the dimension and frequency, wavelength, or equivilent
        limits.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # 1. Apply bandpass filter along time axis from 1 to 100 Hz
    >>> bandpassed = pa.pass_filter(time=(1, 100))

    >>>  # 2. Apply lowpass filter along distance axis for wavelengths less than 100m
    >>> lowpassed = pa.pass_filter(distance=(None, 1/100))
    """
    dim, filt_min, filt_max = _check_filter_kwargs(kwargs)
    axis = patch.dims.index(dim)
    sr = _get_sampling_rate(patch, dim)
    # get niquest and low/high in terms of niquest
    sos = _get_sos(sr, filt_min, filt_max, corners)
    if zerophase:
        out = sosfiltfilt(sos, patch.data, axis=axis)
    else:
        out = sosfilt(sos, patch.data, axis=axis)
    return dascore.Patch(
        data=out, coords=patch.coords, attrs=patch.attrs, dims=patch.dims
    )


@patch_function()
def sobel_filter(patch: PatchType, dim: str, mode="reflect", cval=0.0) -> PatchType:
    """
    Apply a Sobel filter.

    Parameters
    ----------
    **kwargs
        Used to specify the dimension, mode, and for the 'constant' mode, cval.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # 1. Apply Sobel filter using the default parameter values.
    >>> sobel_default = pa.sobel_filter(dim='time', mode='reflect', cval=0.0)

    >>>  # 2. Apply Sobel filter with arbitrary parameter values.
    >>> sobel_arbitrary = pa.sobel_filter(dim='time', mode='constant', cval=1)

    >>> # 3. Apply Sobel filter along both axes
    >>> sobel_time_space = pa.sobel_filter('time',).sobel_filter('distance')
    """
    dim, mode, cval = _check_sobel_args(dim, mode, cval)
    axis = patch.dims.index(dim)
    out = ndimage.sobel(patch.data, axis=axis, mode=mode, cval=cval)
    return dascore.Patch(
        data=out, coords=patch.coords, attrs=patch.attrs, dims=patch.dims
    )


#
# @patch_function()
# def stop_filter(patch: PatchType, corners=4, zerophase=True, **kwargs) -> PatchType:
#     """
#     Apply a Butterworth band stop filter or (highpass, or lowpass).
#
#     Parameters
#     ----------
#     corners
#         The number of corners for the filter.
#     zerophase
#         If True, apply the filter twice.
#     **kwargs
#         Used to specify the dimension and frequency, wavenumber, or equivalent
#         limits.
#
#
#     """
#     dim, filt_min, filt_max = _check_filter_kwargs(kwargs)
#     axis = patch.dims.index(dim)
#     sr = _get_sampling_rate(patch, dim)
#     # get niquest and low/high in terms of niquest
#     sos = _get_sos(sr, filt_min, filt_max, corners)
#     out = sosfilt(sos, patch.data, axis=axis)
#     if zerophase:
#         out = sosfilt(sos, out[::-1])[::-1]
#     return dascore.Patch(data=out, coords=patch.coords, attrs=patch.attrs)


@patch_function()
def median_filter(patch: PatchType, kernel_size=3) -> PatchType:
    """
    Apply 2-D median filter

    Parameters
    ----------
    kernel_size: array_like, optional
        A scalar or a list of length 2, giving the size of the median filter window
        in each dimension. Elements of kernel_size should be odd. If kernel_size is
        a scalar, then this scalar is used as the size in each dimension. Default is
        a kernel of size (3, 3).

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # 1. Apply median filter with 9 time intervals and 5 channels
    >>> filtered_pa = pa.median_filter((9,5))

    Written by Ge Jin (gjin@mines.edu)

    """
    # get niquest and low/high in terms of niquest
    out = medfilt2d(patch.data, kernel_size=kernel_size)
    return dascore.Patch(
        data=out, coords=patch.coords, attrs=patch.attrs, dims=patch.dims
    )



"""
Tests for example fetching.
"""
import pytest

import dascore as dc
from dascore.exceptions import UnknownExample


class TestGetExamplePatch:
    """Test suite for `get_example_patch`"""

    def test_default(self):
        """Ensure calling get_example_patch with no args returns patch."""
        patch = dc.get_example_patch()
        assert isinstance(patch, dc.Patch)

    def test_raises_on_bad_key(self):
        """Ensure a bad key raises expected error."""
        with pytest.raises(UnknownExample, match="No example patch"):
            dc.get_example_patch("NotAnExampleRight????")

    def test_example_1(self):
        """Ensure example 1 returns a Patch"""
        out = dc.get_example_patch("example_event_1")
        assert isinstance(out, dc.Patch)


class TestGetExampleSpool:
    """Test suite for `get_example_spool`"""

    def test_default(self):
        """Ensure calling get_example_spool with no args returns a Spool."""
        patch = dc.get_example_spool()
        assert isinstance(patch, dc.BaseSpool)

    def test_raises_on_bad_key(self):
        """Ensure a bad key raises expected error."""
        with pytest.raises(UnknownExample, match="No example spool"):
            dc.get_example_spool("NotAnExampleRight????")



"""
Tests for reading/writing pickles.
"""
import pytest

import dascore
from dascore.io.pickle.core import PickleIO


@pytest.fixture(scope="session")
def pickle_patch_path(tmp_path_factory, random_patch):
    """Pickle a patch and return the path."""
    path = tmp_path_factory.mktemp("pickle_test") / "test.pkl"
    random_patch.io.write(path, "pickle")
    return path


class TestIsPickle:
    """Test detecting pickle format."""

    def test_detect_file(self, pickle_patch_path):
        """Simple test on output of is pickle."""
        parser = PickleIO()
        out = parser.get_format(pickle_patch_path)
        assert out
        assert out[0] == "PICKLE"

    def test_read_pickle(self, pickle_patch_path, random_patch):
        """Ensure a pickle file can be read."""
        out = dascore.read(pickle_patch_path)
        assert isinstance(out, dascore.BaseSpool)
        assert len(out) == 1
        assert isinstance(out[0], dascore.Patch)
        assert random_patch == out[0]



"""
Tests for reading silixa TDMS file format.
"""

from pathlib import Path

import pytest

import dascore
from dascore.constants import REQUIRED_DAS_ATTRS
from dascore.core.schema import PatchFileSummary
from dascore.io.tdms.core import TDMSFormatterV4713
from dascore.utils.downloader import fetch
from dascore.utils.misc import register_func

TDMS_PATHS = []


@pytest.fixture()
@register_func(TDMS_PATHS)
def TDMS_das_example_path1():
    """Get path for test file"""
    file_path = fetch("sample_tdms_file_v4713.tdms")
    return Path(file_path)


@pytest.fixture()
@register_func(TDMS_PATHS)
def TDMS_das_example_path2():
    """Get path for test file"""
    file_path = fetch("iDAS005_tdms_example.626.tdms")
    return Path(file_path)


@pytest.fixture(params=TDMS_PATHS)
def tdms_das_example_path(request):
    """Get path for test file"""
    return request.getfixturevalue(request.param)


@pytest.fixture()
def tdms_das_patch(tdms_das_example_path):
    """Make patch for test data"""
    ft = TDMSFormatterV4713()
    spool = ft.read(tdms_das_example_path)
    return spool[0]


class TestReadTDMS:
    """Tests for reading the TDMS format."""

    def test_type(self, tdms_das_patch):
        """Ensure the expected type is returned."""
        assert isinstance(tdms_das_patch, dascore.Patch)

    def test_attributes(self, tdms_das_patch):
        """Ensure a few of the expected attrs exist in array."""
        attrs = dict(tdms_das_patch.attrs)
        expected_attrs = {"time_min", "time_max", "distance_min", "data_units"}
        assert set(expected_attrs).issubset(set(attrs))

    def test_has_required_attrs(self, tdms_das_patch):
        """ "Ensure the required das attrs are found"""
        assert set(REQUIRED_DAS_ATTRS).issubset(set(dict(tdms_das_patch.attrs)))

    def test_coord_attr_time_equal(self, tdms_das_patch):
        """The time reported in the attrs and coords should match"""
        attr_time = tdms_das_patch.attrs["time_max"]
        coord_time = tdms_das_patch.coords["time"].max()
        assert attr_time == coord_time

    def test_time_dist_slice(self, tdms_das_patch, tdms_das_example_path):
        """Ensure slicing distance and time works from read func."""
        time_array = tdms_das_patch.coords["time"]
        dist_array = tdms_das_patch.coords["distance"]
        t1, t2 = time_array[10], time_array[40]
        d1, d2 = dist_array[10], dist_array[40]

        patch = TDMSFormatterV4713().read(
            tdms_das_example_path, time=(t1, t2), distance=(d1, d2)
        )[0]
        attrs, coords = patch.attrs, patch.coords
        assert attrs["time_min"] == coords["time"].min() == t1
        assert attrs["time_max"] == coords["time"].max() == t2
        assert attrs["distance_min"] == coords["distance"].min() == d1
        assert attrs["distance_max"] == coords["distance"].max() == d2


class TestGetFormatTDMS:
    """Tests for function to determine if a file is a silixa file."""

    def test_not_silixa_not_tdms(self, dummy_text_file):
        """Test for not a silixa tdms file."""
        parser = TDMSFormatterV4713()
        assert not parser.get_format(dummy_text_file)
        assert not parser.get_format(dummy_text_file.parent)

    def test_silixa_get_format(self, tdms_das_example_path):
        """Test for a silixa tdms file."""
        parser = TDMSFormatterV4713()
        assert parser.get_format(tdms_das_example_path)
        format_name, format_version = parser.get_format(tdms_das_example_path)
        assert format_name == parser.name


class TestScanTDMS:
    """Tests for scanning silixa file."""

    @pytest.fixture
    def tdms_scan(self, tdms_das_example_path):
        """Scan test tdms file, return summary info."""
        out = TDMSFormatterV4713().scan(tdms_das_example_path)
        return out

    def test_scanning(self, tdms_scan):
        """Tests for getting summary info from silixa data."""
        assert isinstance(tdms_scan, list)
        assert len(tdms_scan) == 1
        assert isinstance(tdms_scan[0], PatchFileSummary)

    def test_dims(self, tdms_scan):
        """Ensure dims are populated."""
        for tdms in tdms_scan:
            assert tdms.dims



"""Tests for schema."""
import pydantic
import pytest

from dascore.core.schema import PatchAttrs, SimpleValidator


@pytest.fixture(scope="class")
def random_summary(random_patch) -> PatchAttrs:
    """Return the summary of the random patch."""
    return PatchAttrs.parse_obj(dict(random_patch.attrs))


@pytest.fixture(scope="class")
def random_attrs(random_patch) -> PatchAttrs:
    """Return the summary of the random patch."""
    return random_patch.attrs


class TestSimpleValidator:
    """Test suite for validator."""

    def test_base_validator(self):
        """Ensure the base validator does nothing."""

        class MyModel(pydantic.BaseModel):
            """A test model."""

            value: SimpleValidator = 1

        my_model = MyModel(value=2)
        assert my_model.value == 2


class TestSummarySchema:
    """Tests for summarizing a schema."""

    def test_attrs_reconstructed(self, random_patch, random_summary):
        """Ensure all the expected attrs are extracted."""
        summary1 = random_summary.dict()
        attrs = dict(random_patch.attrs)
        common_keys = set(summary1) & set(attrs)
        for key in common_keys:
            assert summary1[key] == attrs[key]

    def test_can_jsonize(self, random_summary):
        """Ensure the summary can be converted to json."""
        json = random_summary.json()
        assert isinstance(json, str)

    def test_can_roundrip(self, random_summary):
        """Ensure json can be round-tripped"""
        json = random_summary.json()
        random_summary2 = PatchAttrs.parse_raw(json)
        assert random_summary2 == random_summary

    def test_sub_docstrings(self):
        """Ensure the docstring for PatchSummary had params subbed in."""
        docstr = PatchAttrs.__doc__
        # data_type is one of the parameters inserted into docstring.
        assert "data_type" in docstr


class TestSchemaIsDictLike:
    """Tests to insure schema behaves like a dict."""

    def test_get(self, random_attrs):
        """Ensure get returns existing values."""
        out = random_attrs.get("time_min")
        assert out == random_attrs.time_min

    def test_get_existing_key(self, random_attrs):
        """Ensure get returns existing values."""
        out = random_attrs.get("time_min")
        assert out == random_attrs.time_min

    def test_get_no_key(self, random_attrs):
        """Ensure missing keys return default value."""
        out = random_attrs.get("not_a_key", 1)
        assert out == 1

    def test_immutable(self, random_attrs):
        """Ensure random_attrs is faux-immutable."""
        with pytest.raises(TypeError, match="is immutable"):
            random_attrs.bob = 1
        with pytest.raises(TypeError, match="is immutable"):
            random_attrs["bob"] = 1



"""
Tests for transformatter.
"""
import pytest

from dascore.utils.transformatter import FourierTransformatter


@pytest.fixture()
def ft_reformatter():
    """Simple fourier transform formatter."""
    return FourierTransformatter()


class TestFTDimensionRename:
    """Tests for renaming dimensions using FT transformer"""

    dims = ("distance", "time")

    def test_forward_rename_one_index(self, ft_reformatter):
        """Ensure the name can be reassigned."""
        out = ft_reformatter.rename_dims(self.dims, 1)
        assert out == ("distance", "ft_time")

    def test_forward_rename_all_index(self, ft_reformatter):
        """Ensure all indices can be renamed."""
        out = ft_reformatter.rename_dims(self.dims)
        assert out == tuple(f"{ft_reformatter.forward_prefix}{x}" for x in self.dims)

    def test_forward_undo_inverse(self, ft_reformatter):
        """Ensure the inverse is correctly undone."""
        dims = tuple([f"{ft_reformatter.inverse_prefix}{x}" for x in self.dims])
        out = ft_reformatter.rename_dims(dims)
        assert out == self.dims

    def test_inverse_rename_one_index(self, ft_reformatter):
        """Ensure the name can be reassigned."""
        out = ft_reformatter.rename_dims(self.dims, 1, forward=False)
        assert out == ("distance", "ift_time")

    def test_undo_forward_index(self, ft_reformatter):
        """Ensure forward prefex is undone by inverse."""
        dims = tuple(f"{ft_reformatter.forward_prefix}{x}" for x in self.dims)
        out = ft_reformatter.rename_dims(dims, forward=False)
        assert out == self.dims

    def test_double_forward(self, ft_reformatter):
        """Prefixes should stack."""
        pre = ft_reformatter.forward_prefix
        dims1 = ft_reformatter.rename_dims(self.dims)
        dims2 = ft_reformatter.rename_dims(dims1)
        assert dims2 == tuple(f"{pre}{pre}{x}" for x in self.dims)


class TestFTUnitRename:
    """Ensure units can be renamed."""

    dims = ("distance", "time")
    attrs = {"distance_units": "m", "time_units": "s"}

    def test_forward_unit_transform(self, ft_reformatter):
        """Simple forward unit transformation"""
        out = ft_reformatter.rename_attrs(self.dims, self.attrs, index=0)
        key = f"{self.dims[0]}_units"
        value = self.attrs[key]
        assert f"1/({value})" == out[key]
        out_2 = ft_reformatter.rename_attrs(self.dims, out, index=0)
        assert out_2 == self.attrs

    def test_backward_unit_transform(self, ft_reformatter):
        """Simple forward unit transformation"""
        out = ft_reformatter.rename_attrs(self.dims, self.attrs, index=0, forward=False)
        key = f"{self.dims[0]}_units"
        value = self.attrs[key]
        assert f"1/({value})" == out[key]
        out_2 = ft_reformatter.rename_attrs(self.dims, out, index=0, forward=False)
        assert out_2 == self.attrs



"""
Custom filter that makes dynamic (sphinx-esc) cross links work.

Replaces markdown such as [Patch](`dascore.Patch`) with the path to the markdown
file created from dascore/scripts/build_api_docs.py.
"""
import io
import json
import re
import sys
from functools import cache
from pathlib import Path
from typing import Dict


@cache
def get_cross_ref_dict() -> Dict[str, str]:
    """
    Load cross-reference dictionaries.

    Returns
    -------
    """
    out = {}
    path = Path(__file__).absolute()
    while not path.name == "docs":
        path = path.parent
    for cross_ref in path.rglob("cross_ref.json"):
        out.update(json.loads(cross_ref.read_text()))
    return out


def load_stdin(input_stream=None):
    """
    Load input from stdin (json string)
    """
    if input_stream is None:
        input_stream = io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8")
    return input_stream.read()


def dump_stdout(output_str, out_stream=None):
    """
    Dump modified json string to std out.
    """
    if out_stream is None:
        out_stream = sys.stdout
    out_stream.write(output_str)


def replace_links(json_data, raw_string):
    """Replace the aliases (links) with the cross-references."""

    def _yield_links(element):
        """Yield all link elements."""
        if isinstance(element, list):
            for sub_element in element:
                yield from _yield_links(sub_element)
        elif isinstance(element, (dict)):
            if element.get("t") == "Link":
                yield element
            else:
                yield from _yield_links(element.get("c", []))

    reg_pattern = '(?<="\%60)(.*?)(?=\%60")'
    blocks = json_data["blocks"]

    replace_dict = {}
    for link in _yield_links(blocks):
        str_to_scan = json.dumps(link, separators=(",", ":"), ensure_ascii=False)
        if not str_to_scan in raw_string:
            from remote_pdb import RemotePdb

            RemotePdb("127.0.0.1", 4444).set_trace()

        matches = re.finditer(reg_pattern, str_to_scan)
        for match in matches:
            start, stop = match.span()
            cross_refs = get_cross_ref_dict()
            key = str_to_scan[start:stop].replace("`", "")
            new_value = cross_refs.get(key, key)
            if new_value != key:
                new_sub_str = str_to_scan.replace(f'"%60{key}%60"', f'"{new_value}"')
                replace_dict[str_to_scan] = new_sub_str
    for i, v in replace_dict.items():
        raw_string = raw_string.replace(i, v)
    return raw_string


def test():
    """Function to test filter."""
    here = Path(__file__).parent
    data_path = here / "filter_test_data" / "test_data_2.json"
    with data_path.open("r") as fi:
        data = fi.read()
    input_stream = io.StringIO(data)
    main(input_stream)


def main(raw_data=None):
    """Run filter."""
    raw_str = load_stdin(raw_data)
    json_data = json.loads(raw_str)
    output_data = replace_links(json_data, raw_str)
    dump_stdout(output_data)


if __name__ == "__main__":
    # test()  # uncomment to test.
    main()  # uncomment before running quarto so IO works.



"""
IO module for reading Optasense DAS data.
"""
from pathlib import Path
from typing import List, Optional, Union

import dascore as dc
from dascore.constants import timeable_types
from dascore.core.schema import PatchFileSummary
from dascore.io.core import FiberIO
from dascore.utils.hdf5 import HDF5ExtError, NoSuchNodeError, open_hdf5_file

from .utils import _get_qunatx_version_str, _read_quantx, _scan_quantx


class QuantXV2(FiberIO):
    """
    Support for Optasense data format, version 2.
    """

    name = "QUANTX"
    preferred_extensions = ("hdf5", "h5")
    version = "2.0"

    def get_format(self, path: Union[str, Path]) -> Union[tuple[str, str], bool]:
        """
        Return True if file contains Optasense version 2 data else False.

        Parameters
        ----------
        path
            A path to the file which may contain Optasense data.
        """
        try:
            with open_hdf5_file(path, "r") as fi:
                version_str = _get_qunatx_version_str(fi)
                if version_str:
                    return (self.name, version_str)
        except (HDF5ExtError, OSError, IndexError, KeyError, NoSuchNodeError):
            return False

    def scan(self, path: Union[str, Path]) -> List[PatchFileSummary]:
        """
        Scan an Optasense v2 file, return summary information about the file's contents.
        """
        with open_hdf5_file(path) as fi:
            return _scan_quantx(self, fi, path)

    def read(
        self,
        path: Union[str, Path],
        time: Optional[tuple[timeable_types, timeable_types]] = None,
        distance: Optional[tuple[float, float]] = None,
        **kwargs
    ) -> dc.BaseSpool:
        """
        Read an Optasense file.
        """
        # TODO need to create h5 file decorator to avoid too many open/close files.
        with open_hdf5_file(path) as fi:
            patch = _read_quantx(fi.get_node("/Acquisition"), time, distance)
        return dc.spool(patch)



"""
Tests for the file spool.
"""
import pytest

import dascore as dc
from dascore.clients.filespool import FileSpool
from dascore.utils.hdf5 import HDFPatchIndexManager


@pytest.fixture(scope="class")
def terra15_file_spool(terra15_v5_path):
    """A file spool for terra15."""
    return dc.spool(terra15_v5_path)


class TestBasic:
    """Basic tests for the filespool."""

    def test_type(self, terra15_file_spool, terra15_v5_path):
        """Ensure a file spool was returned."""
        assert isinstance(terra15_file_spool, FileSpool)
        assert len(terra15_file_spool) == len(dc.scan_to_df(terra15_v5_path))

    def test_get_patch(self, terra15_file_spool):
        """Ensure the patch is returned."""
        patch = terra15_file_spool[0]
        assert isinstance(patch, dc.Patch)

    def test_str(self, terra15_file_spool):
        """Ensure file spool works."""
        out = str(terra15_file_spool)
        assert "FileSpool" in out

    def test_update(self, tmp_path_factory, random_patch):
        """Update should trigger indexing on formats that support it."""
        path = tmp_path_factory.mktemp("update_test") / "random.h5"
        dc.write(random_patch, path, "dasdae", "1")
        # pre-update
        spool = dc.spool(path)
        contents = spool.get_contents()
        assert not HDFPatchIndexManager(path).has_index
        new_spool = spool.update()
        assert HDFPatchIndexManager(path).has_index
        new_contents = new_spool.get_contents()
        assert contents.equals(new_contents)

    def test_raises_bad_file(self):
        """Simply ensures a bad file will raise."""
        with pytest.raises(FileNotFoundError, match="does not exist"):
            FileSpool("/not/a/directory")



"""Utilities for terra15."""

from typing import Optional

import numpy as np
from tables.exceptions import NoSuchNodeError

from dascore.constants import timeable_types
from dascore.core import Patch
from dascore.core.schema import PatchFileSummary
from dascore.utils.misc import get_slice
from dascore.utils.time import datetime_to_float, to_datetime64

# --- Getting format/version


def _get_terra15_version_str(hdf_fi) -> str:
    """
    Return the version string for terra15 file.
    """

    # define a few root attrs that act as a "fingerprint" for terra15 files
    expected_attrs = [
        "acoustic_bandwidth_end",
        "amplifier_incoming_current",
        "file_start_computer_time",
        "file_version",
    ]
    root_attrs = hdf_fi.root._v_attrs
    is_terra15 = all([hasattr(root_attrs, x) for x in expected_attrs])
    if not is_terra15:
        return ""
    return str(root_attrs.file_version)


# --- Getting File summaries


def _get_scanned_time_min_max(data_node):
    """Get the min/max time from time array."""
    try:
        time = data_node["gps_time"]
    except (NoSuchNodeError, IndexError):
        time = data_node["posix_time"]
    t_len = len(time)
    # first try fast path by tacking first/last of time
    tmin, tmax = time[0], time[-1]
    # This doesn't work if an incomplete datablock exists at the end of
    # the file. In this case we need to read/filter time array (slower).
    if tmin > tmax:
        time = time[:]
        time_filtered = time[time > 0]
        t_len = len(time_filtered)
        tmin, tmax = time_filtered[0], time_filtered[-1]
    return tmin, tmax, t_len


def _get_extra_scan_attrs(self, file_version, path, data_node):
    """Get the extra attributes that go into summary information."""
    tmin, tmax, _ = _get_scanned_time_min_max(data_node)
    out = {
        "time_min": to_datetime64(tmin),
        "time_max": to_datetime64(tmax),
        "path": path,
        "file_format": self.name,
        "file_version": str(file_version),
    }
    return out


def _get_version_data_node(root):
    """Get the version, time, and data node from terra15 file."""
    version = str(root._v_attrs.file_version)
    if version == "4":
        data_type = root._v_attrs.data_product
        data_node = root[data_type]
    elif version in {"5", "6"}:
        data_node = root["data_product"]
    else:
        raise NotImplementedError("Unknown Terra15 version")
    return version, data_node


def _scan_terra15(self, fi, path):
    """Scan a terra15 file, return metadata."""
    root = fi.root
    root_attrs = fi.root._v_attrs
    version, data_node = _get_version_data_node(root)
    out = _get_default_attrs(data_node.data.attrs, root_attrs)
    out.update(_get_extra_scan_attrs(self, version, path, data_node))
    return [PatchFileSummary.parse_obj(out)]


#
# --- Reading patch


def _get_start_stop(time_len, time_lims, file_tmin, dt):
    """Get the start/stop index along time axis."""
    # sst start index
    tmin = time_lims[0] or file_tmin
    tmax = time_lims[1] or dt * (time_len - 1) + file_tmin
    start_ind = int(np.round((tmin - file_tmin) / dt))
    stop_ind = int(np.round((tmax - file_tmin) / dt)) + 1
    # enforce upper limit on time end index.
    if stop_ind > time_len:
        stop_ind = time_len
    assert 0 <= start_ind < stop_ind
    return start_ind, stop_ind


def _get_dar_attrs(data_node, root, tar, dar):
    """Get the attributes for the terra15 data array (loaded)"""
    attrs = _get_default_attrs(data_node.data.attrs, root._v_attrs)
    attrs["time_min"] = tar.min()
    attrs["time_max"] = tar.max()
    attrs["distance_min"] = dar.min()
    attrs["distance_max"] = dar.max()
    return attrs


def _read_terra15(
    root,
    time: Optional[tuple[timeable_types, timeable_types]] = None,
    distance: Optional[tuple[float, float]] = None,
) -> Patch:
    """
    Read a terra15 file.

    Notes
    -----
    The time array is complicated. There is GPS time and Posix time included
    in the file. In version 0.0.6 and less of dascore we just used gps time.
    However, sometimes this results in subsequent samples having a time before
    the previous sample (time did not increase monotonically).

    So now, we use the first GPS sample, then the reported dt to calculate
    time array. The Terra15 folks suggested this is probably the best way to
    do it.
    """
    # get time array
    time_lims = tuple(
        datetime_to_float(x) if x is not None else None
        for x in (time if time is not None else (None, None))
    )
    _, data_node = _get_version_data_node(root)
    file_t_min, file_t_max, time_len = _get_scanned_time_min_max(data_node)
    # surprisingly, using gps time column, dt is much different than dt
    # reported in data attrs!, use GPS time here.
    dt = (file_t_max - file_t_min) / (time_len - 1)
    # get the start and stop along the time axis
    start_ind, stop_ind = _get_start_stop(time_len, time_lims, file_t_min, dt)
    req_t_min = file_t_min if start_ind == 0 else file_t_min + dt * start_ind
    # account for files that might not be full, adjust requested max time
    stop_ind = min(stop_ind, time_len)
    assert stop_ind > start_ind
    req_t_max = (
        time_lims[-1] if stop_ind < time_len else file_t_min + (stop_ind - 1) * dt
    )
    assert req_t_max > req_t_min
    # calculate time array, convert to datetime64
    t_float = file_t_min + np.arange(start_ind, stop_ind) * dt
    time_ar = to_datetime64(t_float)
    time_inds = (start_ind, stop_ind)
    # get data and sliced distance coord
    dist_ar = _get_distance_array(root)
    dslice = get_slice(dist_ar, distance)
    dist_ar_trimmed = dist_ar[dslice]
    data = data_node.data[slice(*time_inds), dslice]
    coords = {"time": time_ar, "distance": dist_ar_trimmed}
    dims = ("time", "distance")
    attrs = _get_dar_attrs(data_node, root, time_ar, dist_ar_trimmed)
    return Patch(data=data, coords=coords, attrs=attrs, dims=dims)


def _get_default_attrs(data_node_attrs, root_node_attrs):
    """
    Return the required/default attributes which can be fetched from attributes.

    Note: missing time, distance absolute ranges. Downstream functions should handle
    this.
    """
    out = dict(dims="time, distance")
    _root_attrs = {
        "data_product": "data_type",
        "dx": "d_distance",
        "serial_number": "instrument_id",
        "sensing_range_start": "distance_min",
        "sensing_range_end": "distance_max",
        "data_product_units": "data_units",
    }
    for treble_name, out_name in _root_attrs.items():
        out[out_name] = getattr(root_node_attrs, treble_name)

    out["d_time"] = data_node_attrs.dT

    return out


def _get_distance_array(root):
    """Get the distance (along fiber) array."""
    # TODO: At least for the v4 test file, sensing_range_start, sensing_range_stop,
    # nx, and dx are not consistent, meaning d_min + dx * nx != d_max
    # so I just used this method. We need to look more into this.
    attrs = root._v_attrs
    dist = (np.arange(attrs.nx) * attrs.dx) + attrs.sensing_range_start
    return dist



"""Module for reporting the version of dascore."""
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("dascore")
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"

__last_version__ = ".".join(__version__.split(".")[:3])



"""
Tests for time variables.
"""
import functools

import numpy as np
import pandas as pd
import pytest

from dascore.utils.time import (
    get_select_time,
    is_datetime64,
    to_datetime64,
    to_number,
    to_timedelta64,
)


class TestToDateTime64:
    """Tests for converting things to datetime64."""

    date_strs = ["1970-01-01", "2020-01-03T05:22:11.123123345", "2017-09-18T01"]

    def test_float_array(self):
        """Ensure basic tests work"""
        date_strs = ["2015-01-01", "2020-03-01T21:10:10", "1970-01-02"]
        input_datetime64 = np.array(date_strs, dtype="datetime64[ns]")
        # convert to float (ns) then divide by e9 to get float in seconds
        float_array = input_datetime64.astype(np.float64) / 1_000_000_000
        out = to_datetime64(float_array)
        assert np.all(input_datetime64 == out)

    def test_single_float(self):
        """Ensure a single float can be converted to datetime."""
        out = to_datetime64(1.0)
        assert isinstance(out, np.datetime64)
        assert out == np.datetime64("1970-01-01T00:00:01.00000", "ns")

    def test_string(self):
        """Test for converting a string to a datetime object"""
        for time_str in self.date_strs:
            out = to_datetime64(time_str)
            assert isinstance(out, np.datetime64)
            assert time_str in str(out)

    def test_str_array(self):
        """Tests for converting an array of strings."""
        out = to_datetime64(self.date_strs)
        assert len(out) == len(self.date_strs)
        for el, datestr in zip(out, self.date_strs):
            assert datestr in str(el)

    def test_datetime64_array(self):
        """Tests for inputting datetime64."""
        array = to_datetime64(self.date_strs)
        out = to_datetime64(array)
        for el, datestr in zip(out, self.date_strs):
            assert datestr in str(el)

    def test_datetime64(self):
        """A datetitme64 should remain thus and equal."""
        d_time = to_datetime64("2020-01-01")
        out = to_datetime64(d_time)
        assert d_time == out

    def test_none(self):
        """None should return NaT."""
        out = to_datetime64(None)
        assert pd.isnull(out)

    def test_pandas_timestamp(self):
        """Ensure a timestamp returns the datetime64."""
        ts = pd.Timestamp("2020-01-03")
        expected = ts.to_datetime64()
        out = to_datetime64(ts)
        assert out == expected

    def test_non_ns_datetime64(self):
        """Tests that a non-nano second datatime gets converted to one."""
        datetimes = [
            np.datetime64("2011-01-01", "s"),
            np.datetime64("2011-01-01", "ms"),
            np.datetime64("2011-01-01", "ns"),
        ]
        expected = np.datetime64("2011-01-01", "ns")
        for dt in datetimes:
            out = to_datetime64(dt)
            assert out == expected
            # check string rep to ensure precision matches
            assert str(out) == str(expected)

    def test_array_with_all_nan(self):
        """Tests for NaN in array."""
        array = np.array([None, None])
        out = to_datetime64(array)
        assert pd.isnull(out).all()

    def test_array_with_one_nan(self):
        """Tests for NaN in array."""
        array = np.array([None, 1.2])
        out = to_datetime64(array)
        assert pd.isnull(out[0]) and not pd.isnull(out[1])

    def test_one_nan_one_timestamp(self):
        """Ensure (None, TimeStamp) works."""
        array = (None, pd.Timestamp("2020-01-01"))
        out = to_datetime64(array)
        assert len(out) == 2
        assert pd.isnull(out[0])
        assert out[1] == array[1]

    def test_str_tuples(self):
        """Ensure tuples of datetime strings can also be converted."""
        out1 = to_datetime64((None, "2011-01-01"))
        out2 = to_datetime64(("2011-01-01", None))
        assert pd.isnull(out2[1]) and pd.isnull(out1[0])
        assert out1[1] == out2[0]

    def test_zero_dimensional_array(self):
        """
        Ensure a zero-dimensional array of a datetime is just unpacked.
        """
        ar = np.array("2011-08-12", dtype="datetime64[ns]")
        out = to_datetime64(ar)
        assert isinstance(out, np.datetime64)

    def test_negative_float(self):
        """Negative floats should be a symmetric operation."""
        floats_to_test = [0.1, 10.0, 10.001, 100.0, 0, np.inf]
        for val in floats_to_test:
            out = to_datetime64(-val)
            expected = to_datetime64(0) - to_timedelta64(val)
            assert out == expected or (pd.isnull(out) and pd.isnull(expected))

    def test_negative_int(self):
        """Negative ints should be a symmetric operation."""
        floats_to_test = [0, 10, 100]
        for val in floats_to_test:
            out = to_datetime64(-val)
            expected = to_datetime64(0) - to_timedelta64(val)
            assert out == expected or (pd.isnull(out) and pd.isnull(expected))


class TestToTimeDelta64:
    """Tests for creating timedeltas"""

    @pytest.fixture(
        params=(
            np.timedelta64(1, "ns"),
            np.timedelta64(10, "s"),
            np.timedelta64(63, "ms"),
        )
    )
    def timedelta64(self, request):
        """Return the parametrized timedeltas"""
        return request.param

    def test_single_float(self):
        """Ensure a single float is converted to timedelta"""
        out = to_timedelta64(1.0)
        assert out == np.timedelta64(1_000_000_000, "ns")

    def test_float_array(self):
        """Ensure an array of flaots can be converted to ns timedelta"""
        ar = [1.0, 0.000000001, 0.001]
        expected = np.array([1 * 10**9, 1, 1 * 10**6], "timedelta64")
        out = to_timedelta64(ar)
        assert np.all(out == expected)

    def test_timedelta64_array(self):
        """Ensure passing timedelta array works."""
        expected = np.array([1 * 10**9, 1, 1 * 10**6], "timedelta64[s]").astype(
            "timedelta64[ns]"
        )
        out = to_timedelta64(expected)
        assert np.equal(out, expected).all()

    def test_timedetla64(self):
        """Test for passing a time delta."""
        td = to_timedelta64(123)
        out = np.timedelta64(123, "s")
        out2 = to_timedelta64(out)
        assert out == td == out2

    def test_pandas_time_delta(self):
        """Ensure pandas timedelta still works"""
        expected = np.timedelta64(1, "s")
        ptd = pd.Timedelta(1, "s")
        out = to_timedelta64(ptd)
        assert expected == out

    def test_str_roundtrip(self, timedelta64):
        """Ensure the output of str(timedelta64) is valid input."""
        obj_str = str(timedelta64)
        assert timedelta64 == to_timedelta64(obj_str)

    def test_array_with_all_nan(self):
        """Tests for NaN in array."""
        array = np.array([None, None])
        out = to_timedelta64(array)
        assert pd.isnull(out).all()

    def test_array_with_one_nan(self):
        """Tests for NaN in array."""
        array = np.array([None, 1.2])
        out = to_timedelta64(array)
        assert pd.isnull(out[0]) and not pd.isnull(out[1])

    def test_zero_dimensional_array(self):
        """A degenerate array should be unpacked."""
        array1 = np.array("1", dtype="timedelta64[s]")
        assert isinstance(to_timedelta64(array1), np.timedelta64)
        array2 = np.array(1, dtype="timedelta64[s]")
        assert isinstance(to_timedelta64(array2), np.timedelta64)

    def test_negative_float(self):
        """Negative floats should be a symmetric operation."""
        floats_to_test = [0.1, 10.0, 10.001, 100.0, 0, np.inf]
        for val in floats_to_test:
            out = to_timedelta64(-val)
            expected = -to_timedelta64(abs(val))
            assert out == expected or (pd.isnull(out) and pd.isnull(expected))

    def test_negative_int(self):
        """Negative ints should be a symmetric operation."""
        floats_to_test = [1, 10, 100]
        for val in floats_to_test:
            out = to_timedelta64(-val)
            expected = -to_timedelta64(abs(val))
            assert out == expected or (pd.isnull(out) and pd.isnull(expected))


class TestGetSelectTime:
    """Tests for getting select time(s)."""

    t1 = np.datetime64("2020-01-03")
    t2 = np.datetime64("2020-01-04")
    func = functools.partial(get_select_time, time_min=t1, time_max=t2)

    def test_datetime64(self):
        """Test for returning a datetime 64."""
        assert get_select_time(self.t1) == self.t1
        assert get_select_time(self.t2) == self.t2
        time = self.t1 + np.timedelta64(100, "s")
        assert get_select_time(time) == time

    def test_string(self):
        """Tests for passing time as a string"""
        tstr = "2020-01-03T01"
        out = self.func(tstr)
        assert out == np.datetime64(tstr)

    def test_positive_float_int(self):
        """Test float/ints are interpreted as relative"""
        out = self.func(1)
        expected = self.t1 + to_timedelta64(1)
        assert out == expected

        out = self.func(1.0)
        expected = self.t1 + to_timedelta64(1)
        assert out == expected

    def test_negative_float_int(self):
        """Test float/ints are interpreted as relative"""
        out = self.func(-1)
        expected = self.t2 - to_timedelta64(1)
        assert out == expected

        out = self.func(-1.0)
        expected = self.t2 - to_timedelta64(1)
        assert out == expected


class TestToNumber:
    """Tests for converting time-like types to ints, or passing through reals"""

    def test_timedelta64(self):
        """Ensure a timedelta64 returns the ns"""
        out = to_number(np.timedelta64(1, "s"))
        assert out == 1_000_000_000

    def test_datetime64(self):
        """Ensure int ns is returned for datetime64."""
        out = to_number(to_datetime64("1970-01-01") + np.timedelta64(1, "ns"))
        assert out == 1

    def test_timedelta64_array(self):
        """Ensure int ns is returned for datetime64."""
        array = to_datetime64(["2017-01-01", "1970-01-01", "1999-01-01"])
        out = to_number(array)
        assert np.issubdtype(out.dtype, np.int64)

    def test_timedelta_array(self):
        """Ensure a timedelta array works."""
        array = to_timedelta64([1, 1_000_000, 20])
        out = to_number(array)
        assert np.issubdtype(out.dtype, np.int64)

    def test_nullish_returns_nan(self):
        """Ensure a timedelta array works."""
        assert to_number(None) is np.NaN
        assert to_number(pd.NaT) is np.NaN

    def test_number_unchanged(self):
        """Ensure a number is passed through unchanged."""
        assert to_number(10) == 10
        assert to_number(10.1) == 10.1

    def test_numeric_array_unchanged(self):
        """Ensure numeric arrays are not changed."""
        array = np.array([10, 12, 20])
        assert np.all(to_number(array) == array)
        array = np.array([12.3, 13.2, 12.2])
        assert np.all(to_number(array) == array)

    def test_non_ns_datetime64(self):
        """Tests that a non-nano second datatime gets converted to one."""
        datetimes = [
            np.datetime64("2011-01-01", "s"),
            np.datetime64("2011-01-01", "ms"),
            np.datetime64("2011-01-01", "ns"),
        ]
        expected = np.datetime64("2011-01-01", "ns").astype(np.int64)
        for dt in datetimes:
            out = to_number(dt)
            assert out == expected


class TestIsDateTime:
    """Ensure is_datetime64 detects datetimes."""

    def test_not_datetime(self):
        """Simple tests for things that aren't datetimes."""
        assert not is_datetime64(None)
        assert not is_datetime64(float)
        assert not is_datetime64(10)
        assert not is_datetime64(42.12)
        assert not is_datetime64(np.timedelta64(10, "s"))

    def test_is_datetime(self):
        """Things that should return True"""
        assert is_datetime64(np.datetime64("1970-01-01"))
        array = to_datetime64(["1990-01-01", "2010-01-01T12:23:22"])
        assert is_datetime64(array)

    def test_datetime_series(
        self,
    ):
        """is_datettime should work with a pandas series."""
        array = to_datetime64(["1990-01-01", "2010-01-01T12:23:22"])
        ser = pd.Series(array)
        assert is_datetime64(ser)



"""
Module for re-sampling patches.
"""
from typing import Literal, Union

import numpy as np
from scipy.signal import decimate as scipy_decimate

import dascore as dc
import dascore.compat as compat
from dascore.constants import PatchType
from dascore.exceptions import FilterValueError
from dascore.utils.misc import check_evenly_sampled
from dascore.utils.patch import (
    get_dim_value_from_kwargs,
    get_start_stop_step,
    patch_function,
)
from dascore.utils.time import to_number, to_timedelta64


@patch_function()
def decimate(
    patch: PatchType,
    filter_type: Literal["iir", "fir", None] = "iir",
    copy=True,
    **kwargs,
) -> PatchType:
    """
    Decimate a patch along a dimension.

    Parameters
    ----------
    filter_type
        filter type to use to avoid aliasing. Options are:
            iir - infinite impulse response
            fir - finite impulse response
            None - No pre-filtering
    copy
        If True, copy the decimated data array. This is needed if you want
        the old array to get gc'ed to free memory otherwise a view is returned.
        Only applies when filter_type == None.
    **kwargs
        Used to pass dimension and factor. For example `time=10` is 10x
        decimation along the time axis.

    Notes
    -----
    Simply uses scipy.signal.decimate if filter_type is specified. Otherwise,
    just slice data long specified dimension only including every n samples.

    Examples
    --------
    # Simple example using iir
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> decimated_irr = patch.decimate(time=10, filter_type='iir')
    >>> # Example using fir along distance dimension
    >>> decimated_fir = patch.decimate(distance=10, filter_type='fir')
    """
    dim, axis, factor = get_dim_value_from_kwargs(patch, kwargs)
    # Apply scipy.signal.decimate and geet new coords
    if filter_type:
        if filter_type == "IRR" and factor > 13:
            msg = (
                "IRR filter is unstable for decimation factors above"
                " 13. Call decimate multiple times."
            )
            raise FilterValueError(msg)
        data = scipy_decimate(patch.data, factor, ftype=filter_type, axis=axis)
        coords = {x: patch.coords[x] for x in patch.dims}
        coords[dim] = coords[dim][::factor]
    else:  # No filter, simply slice along specified dimension.
        dar = patch._data_array.sel(**{dim: slice(None, None, factor)})
        # Need to copy so array isn't a slice and holds onto reference of parent
        data = dar.data if not copy else dar.data.copy()
        coords = dar.coords
    # Update delta_dim since spacing along dimension has changed.
    attrs = dict(patch.attrs)
    attrs[f"d_{dim}"] = patch.attrs[f"d_{dim}"] * factor
    out = dc.Patch(data=data, coords=coords, attrs=attrs, dims=patch.dims)
    return out


@patch_function()
def interpolate(
    patch: PatchType, kind: Union[str, int] = "linear", **kwargs
) -> PatchType:
    """
    Set coordinates of patch along a dimension using interpolation.

    Parameters
    ----------
    patch
        The patch object to which interpolation is applied.
    kind
        The type of interpolation. See Notes for more details.
        If a string, the following are supported:
            linear - linear interpolation between a pair of points.
            nearest - use the nearest sample for interpolation.
        If an int, it specifies the order of spline to use. EG 1 is a linear
            spline, 2 is quadratic, 3 is cubic, etc.

    **kwargs
        Used to specify dimension and interpolation values.

    Notes
    -----
    This function just uses scipy's interp1d function under the hood.
    See scipy.interpolate.interp1d for information.

    Values for interpolation must be evenly-spaced.
    """
    dim, axis, samples = get_dim_value_from_kwargs(patch, kwargs)
    # ensure samples are evenly sampled
    check_evenly_sampled(samples)
    # we need to make sure only real numbers are used, interp1d doesn't support
    # datetime64 yet.
    coord_num = to_number(patch.coords[dim])
    samples_num = to_number(samples)
    func = compat.interp1d(coord_num, patch.data, axis=axis, kind=kind)
    out = func(samples_num)
    # update attributes
    new_attrs = dict(patch.attrs)
    new_attrs[f"d_{dim}"] = np.median(np.diff(samples))
    new_attrs[f"min_{dim}"] = np.min(samples)
    new_attrs[f"max_{dim}"] = np.max(samples)
    # update coordinates
    new_coords = {x: patch.coords[x] for x in patch.dims}
    new_coords[dim] = samples
    kwargs = dict(data=out, attrs=new_attrs, dims=patch.dims, coords=new_coords)
    return patch.__class__(**kwargs)


@patch_function()
def resample(
    patch: PatchType, window=None, interp_kind="linear", **kwargs
) -> PatchType:
    """
    Resample along a single dimension using Fourier Method and interpolation.

    The dimension which should be resampled is passed as kwargs. The key
    is the dimension name and the value is the new sampling period.

    Since Fourier methods only support adding or removing an integer number
    of frequency bins, the exact desired sampling rate is often not achievable
    with resampling alone. If the fourier resampling doesn't produce the exact
    an interpolation (see [interpolate](`dascore.proc.interpolate`)) is used to
    achieve the desired sampling rate.

    Parameters
    ----------
    patch
        The patch to resample.
    window
        The Fourier-domain window that tapers the Fourier spectrum. See
        scipy.signal.resample for details. Only used if method == 'fft'.
    interp_kind
        The interpolation type if output of fourier resampling doesn't produce
        exactly the right sampling rate.
    **kwargs
        keyword arguments to specify

    Notes
    -----
    Unlike [iresample](`dascore.proc.iresample`) this function requires a
    sampling_period.

    Often the resulting Patch will be slightly shorter than the input Patch.

    Examples
    --------
    # resample a patch along time dimension to 10 ms
    import dascore as dc
    patch = dc.get_example_patch()
    new = patch.resample(time=np.timedelta64(10, 'ms'))

    See Also
    --------
    [decimate](`dascore.proc.resample.decimate`)
    [interpolate](`dascore.proc.resample.interpolate`)
    [iresample](`dascore.proc.resmaple.irsample`)
    """
    dim, axis, new_d_dim = get_dim_value_from_kwargs(patch, kwargs)
    d_dim = patch.attrs[f"d_{dim}"]  # current sampling rate
    # nasty hack so that ints/floats get converted to seconds.
    if isinstance(d_dim, np.timedelta64):
        new_d_dim = to_timedelta64(new_d_dim)
    # dim_range = int(compat.floor((dim_stop - dim_start) / val))
    current_sig_len = patch.data.shape[axis]
    new_len = current_sig_len * (d_dim / new_d_dim)
    out = iresample.func(patch, window=window, **{dim: int(np.round(new_len))})
    # Interpolate if new sampling rate is not very close to desired sampling rate.
    if not np.isclose(new_len, np.round(new_len)):
        start, stop, step = get_start_stop_step(out, dim)
        new_coord = np.arange(start, stop, new_d_dim)
        out = interpolate(out, kind=interp_kind, **{dim: new_coord})
    return out


@patch_function()
def iresample(patch: PatchType, window=None, **kwargs) -> PatchType:
    """
    Resample a patch along a single dimension using Fourier Method.

    Unlike [resample](`dascore.proc.resample`) this function requires the
    number of samples for the selected dimension.

    Parameters
    ----------
    patch
        The patch to resample.
    window
        The Fourier-domain window that tapers the Fourier spectrum. See
        scipy.signal.resample for details.
    **kwargs
         keyword arguments to specify dimension.

    Notes
    -----
    Simply uses scipy.signal.resample.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> new = patch.iresample(time=50)

    See Also
    --------
    [iresample](`dascore.proc.resmaple.irsample`)
    [decimate](`dascore.proc.resample.decimate`)
    """
    dim, axis, new_length = get_dim_value_from_kwargs(patch, kwargs)
    coord = patch.coords[dim]
    out, new_coord = compat.resample(
        patch.data, new_length, t=coord, axis=axis, window=window
    )
    # update coordinates
    new_coords = {x: patch.coords[x] for x in patch.dims}
    new_coords[dim] = new_coord
    # update attributes
    new_attrs = dict(patch.attrs)
    new_attrs[f"d_{dim}"] = new_coord[1] - new_coord[0]
    new_attrs[f"{dim}_max"] = np.max(new_coord)
    patch_inputs = dict(data=out, attrs=new_attrs, dims=patch.dims, coords=new_coords)
    return patch.__class__(**patch_inputs)



"""
Tests for FileSpool.
"""
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.clients.dirspool import DirectorySpool
from dascore.constants import ONE_SECOND
from dascore.core.schema import PatchFileSummary
from dascore.utils.hdf5 import HDFPatchIndexManager
from dascore.utils.misc import register_func

DIRECTORY_SPOOLS = []


@pytest.fixture(scope="class")
@register_func(DIRECTORY_SPOOLS)
def one_directory_spool(one_file_dir):
    """Create a directory with a single DAS file."""
    spool = DirectorySpool(one_file_dir)
    return spool.update()


@pytest.fixture(scope="class", params=DIRECTORY_SPOOLS)
def directory_spool(request):
    """Meta fixture for getting all file spools."""
    return request.getfixturevalue(request.param)


class TestFileSpool:
    """Test that the file spool works."""

    def test_isinstance(self, directory_spool):
        """Simply ensure expected type was returned."""
        assert isinstance(directory_spool, DirectorySpool)


class TestDirectoryIndex:
    """Tests for returning summaries of all files in managed directory."""

    @pytest.fixture(scope="class")
    def basic_index_df(self, basic_file_spool):
        """Return the index file of the basic bank."""
        spool = basic_file_spool
        return spool.get_contents()

    def test_index_exists(self, basic_file_spool):
        """An index should be returned."""
        assert basic_file_spool.indexer.index_path.exists()

    def test_index_len(self, basic_index_df, two_patch_directory):
        """An index should be returned."""
        bank_paths = list(Path(two_patch_directory).rglob("*hdf5"))
        assert isinstance(basic_index_df, pd.DataFrame)
        assert len(bank_paths) == len(basic_index_df)

    def test_index_columns(self, basic_index_df):
        """Ensure expected columns show up in the index."""
        schema_fields = PatchFileSummary.get_index_columns()
        assert set(basic_index_df).issuperset(schema_fields)

    def test_patches_extracted(self, basic_file_spool):
        """Ensure the patches can be extracted."""
        index = basic_file_spool.get_contents()
        patches = [x for x in basic_file_spool]
        assert len(index) == len(patches)
        for patch in patches:
            assert isinstance(patch, dc.Patch)

    def test_str_(self, basic_file_spool):
        """Ensure the filespool has a useful str/repr."""
        out = str(basic_file_spool)
        # ensure the default str is not used.
        assert "object at" not in out

    def test_long_tags(self, random_patch, tmp_path):
        """Ensure a long tag still works."""
        new = random_patch.update_attrs(tag="hey" * 15)
        path = tmp_path / "test.h5"
        new.io.write(path, "dasdae")
        spool = dc.spool(path).update()
        isinstance(spool, dc.BaseSpool)

    def test_specify_index_path(self, random_patch, tmp_path_factory):
        """
        Ensure an external path can be specified for the index. See #129.
        """
        bank_path = tmp_path_factory.mktemp("bank")
        index_path = tmp_path_factory.mktemp("index") / "index.h5"
        random_patch.io.write(bank_path / "contents.h5", "dasdae")
        spool1 = dc.spool(bank_path, index_path=index_path)
        spool1.update()
        # ensure the index was created in the expected place
        assert spool1.indexer.index_path == index_path
        # ensure the default index file was not written
        spool2 = dc.spool(bank_path)
        default_index_path = spool2.indexer.index_path
        assert not default_index_path.exists()
        # next ensure the index path is used
        spool3 = dc.spool(bank_path, index_path=index_path)
        df = spool3.get_contents()
        assert len(df) == 1
        patch = spool3[0]
        assert isinstance(patch, dc.Patch)
        assert not default_index_path.exists()


class TestSelect:
    """tests for subselecting data."""

    @pytest.fixture(scope="class")
    def spool_tag(self, basic_file_spool):
        """Return a string of a tag in the basic_file_spool."""
        contents = basic_file_spool.get_contents()
        tag = contents.loc[contents["tag"].astype(bool), "tag"].iloc[0]
        return tag

    def test_subselect_trims_patches(self, basic_file_spool):
        """Ensure sub-selecting trims start/end times on df and output patches."""
        current = basic_file_spool.get_contents()
        new_min = current["time_min"].min() + np.timedelta64(2, "s")
        new_max = current["time_max"].max() - np.timedelta64(2, "s")
        spool = basic_file_spool.select(time=(new_min, new_max))
        # the limits of rows which were intersected should have been trimmed.
        df = spool.get_contents()
        assert (df["time_min"] >= new_min).all()
        assert (df["time_max"] <= new_max).all()
        # as well as the patches produced
        for patch in spool:
            assert patch.attrs["time_min"] >= new_min
            assert patch.attrs["time_max"] <= new_max

    def test_sub_select_tag_equals(self, basic_file_spool, spool_tag):
        """Ensure selecting stations works."""
        new = basic_file_spool.select(tag=spool_tag)
        new_contents = new.get_contents()
        assert (new_contents["tag"] == spool_tag).all()

    def test_is_in_tag(self, basic_file_spool, spool_tag):
        """Ensure tags can also be selected from a collection."""
        tag_collection = {spool_tag, "bob", "bill"}
        out = basic_file_spool.select(tag=tag_collection).get_contents()
        assert out["tag"].isin(tag_collection).all()

    def test_multiple_selects(self, diverse_directory_spool):
        """Ensure selects can be stacked."""
        spool = diverse_directory_spool
        contents = spool.get_contents()
        duration = contents["time_max"] - contents["time_min"]
        new_max = (contents["time_min"] + duration.mean() / 2).median()
        out = (
            spool.select(network="das2").select(tag="ran*").select(time=(None, new_max))
        )
        assert len(out) > 0
        # first check content dataframe
        new_content = out.get_contents()
        assert len(new_content) == len(out)
        assert (new_content["network"] == "das2").all()
        assert (new_content["tag"].str.startswith("ran")).all()
        assert (new_content["time_max"] <= new_max).all()
        # then check patches
        for patch in out:
            attrs = patch.attrs
            assert attrs["network"] == "das2"
            assert attrs["tag"].startswith("ran")
            assert attrs["time_max"] <= new_max

    def test_select_time_tuple_with_string(self, basic_file_spool):
        """Ensure time tuples with strings still work."""
        time_str = "2017-09-18T00:00:04"
        dt = np.datetime64(time_str)
        spool1 = basic_file_spool.select(time=(None, dt))
        spool2 = basic_file_spool.select(time=(None, time_str))
        for pa1, pa2 in zip(spool1, spool2):
            assert pa1.attrs["time_max"] == pa2.attrs["time_max"]

    def test_select_non_zero_index(self, diverse_directory_spool):
        """
        A Bug caused the contents of the source dataframe to have
        non-zero based indices, thus spools didnt work. This fixes
        the issue.
        """
        contents = diverse_directory_spool.get_contents()
        end_time = contents["time_max"].min()
        sub = diverse_directory_spool.select(
            time=(None, end_time),
            distance=(100, 200),
        )
        assert len(sub) == 1
        patch = sub[0]
        assert isinstance(patch, dc.Patch)


class TestBasicChunk:
    """Tests for chunking filespool."""

    def test_directory_path_doesnt_change(self, one_file_directory_spool):
        """Chunking shouldn't change the path to the managed directory."""
        out = one_file_directory_spool.chunk(time=1)
        assert out.spool_path == one_file_directory_spool.spool_path

    def test_chunk_doesnt_modify_original(self, one_file_directory_spool):
        """Chunking shouldn't modify original spool or its dfs."""
        spool = one_file_directory_spool
        contents_before_chunk = spool.get_contents()
        _ = spool.chunk(time=2)
        contents_after_chunk = spool.get_contents()
        assert contents_before_chunk.equals(contents_after_chunk)

    def test_sub_chunk(self, one_file_directory_spool):
        """Ensure the patches can be subdivided."""
        spool = one_file_directory_spool
        contents = spool.get_contents()
        durations = contents["time_max"] - contents["time_min"]
        new_t_delta = (durations / 4).max()
        new_spool = spool.chunk(time=new_t_delta, keep_partial=True)
        # Ensure there are exactly 4x as many patches in spool after chunk
        new_contents = new_spool.get_contents()
        assert len(new_contents) == 4 * len(spool)
        # Ensure each spool can be iterated
        patch_list = list(new_spool)
        for patch in patch_list:
            assert isinstance(patch, dc.Patch)


class TestGetContents:
    """Tests for getting the contents of the spool."""

    def test_str_columns_in_dataframe(self, diverse_directory_spool):
        """Ensure all the string columns are in index."""
        df = diverse_directory_spool.get_contents()
        expected = HDFPatchIndexManager._min_itemsize
        assert set(df.columns).issuperset(set(expected))


class TestFileSpoolIntegrations:
    """Small integration tests for the file spool."""

    def test_one(self, diverse_spool_directory):
        """Small integration test with diverse spool."""
        network = "das2"
        endtime = np.datetime64("2022-01-01")
        duration = 3
        spool = (
            dc.spool(diverse_spool_directory)
            .select(network=network)  # sub-select das2 network
            .select(time=(None, endtime))  # unselect anything after 2022
            .chunk(time=duration, overlap=0.5)  # change the chunking of the patches
        )
        for patch in spool:
            assert isinstance(patch, dc.Patch)
            attrs = patch.attrs
            assert attrs["network"] == network
            assert attrs["time_max"] <= endtime
            patch_duration = (attrs["time_max"] - attrs["time_min"]) / ONE_SECOND
            diff = patch_duration - duration
            assert abs(diff) <= 1.5 * attrs["d_time"] / ONE_SECOND



"""
Utilities for working with the Patch class.
"""
import collections
import copy
import functools
import inspect
import re
import warnings
from fnmatch import translate
from typing import (
    Any,
    Callable,
    Dict,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import xarray as xr

import dascore as dc
from dascore.constants import PATCH_MERGE_ATTRS, PatchType, SpoolType
from dascore.core.schema import PatchAttrs, PatchFileSummary
from dascore.exceptions import PatchAttributeError, PatchDimError
from dascore.utils.coords import Coords
from dascore.utils.docs import compose_docstring, format_dtypes
from dascore.utils.misc import append_func
from dascore.utils.time import to_datetime64, to_timedelta64

attr_type = Union[Dict[str, Any], str, Sequence[str], None]


def _shallow_copy(patch: PatchType) -> PatchType:
    """
    Shallow copy patch so data array, attrs, and history can be changed.

    Note
    ----
    This is an internal function because Patch should be immutable with
    the public APIs.
    """
    dar = patch._data_array.copy(deep=False)  # dont copy data and such
    attrs = dict(dar.attrs)
    attrs["history"] = list(attrs.get("history", []))
    dar.attrs = attrs
    return patch.__class__(dar)


def _func_and_kwargs_str(func: Callable, patch, *args, **kwargs) -> str:
    """
    Get a str rep of the function and input args.
    """
    callargs = inspect.getcallargs(func, patch, *args, **kwargs)
    callargs.pop("patch", None)
    callargs.pop("self", None)
    kwargs_ = callargs.pop("kwargs", {})
    arguments = []
    arguments += [f"{k}={repr(v)}" for k, v in callargs.items() if v is not None]
    arguments += [f"{k}={repr(v)}" for k, v in kwargs_.items() if v is not None]
    arguments.sort()
    out = f"{func.__name__}("
    if arguments:
        out += f"{','.join(arguments)}"
    return out + ")"


def _get_history_str(patch: PatchType, func, *args, _history="full", **kwargs) -> str:
    """
    Log history of a function being called on a Patch object.

    Parameters
    ----------
    patch
        The patch which will track history.
    func
        A callable which takes the patch as the first argument.
    *args
        The arguments passed to the function.
    _history
        String specifying how the history is to be recorded.
    **kwargs
        kwargs for func.
    """
    if _history is None:
        return None
    if _history == "full":
        history_str = _func_and_kwargs_str(func, patch, *args, **kwargs)
    else:
        history_str = str(func.__name__)
    return history_str


def check_patch_dims(
    patch: PatchType, required_dims: Optional[Tuple[str, ...]]
) -> PatchType:
    """
    Check if a patch object has required dimensions, else raise PatchDimError.

    Parameters
    ----------
    patch
        The input patch
    required_dims
        A tuple of required dimensions.
    """
    if required_dims is None:
        return patch
    dim_set = set(patch.dims)
    missing_dims = set(required_dims) - set(patch.dims)
    if not set(required_dims).issubset(dim_set):
        msg = f"patch is missing required dims: {tuple(missing_dims)}"
        raise PatchDimError(msg)
    return patch


def check_patch_attrs(patch: PatchType, required_attrs: attr_type) -> PatchType:
    """
    Check for expected attributes.

    Parameters
    ----------
    patch
        The patch to validate
    required_attrs
        The expected attrs. Can be a sequence or mapping. If sequence, only
        check that attrs exist. If mapping also check that values are equal.
    """
    if required_attrs is None:
        return patch
    # test that patch attr mapping is equal
    patch_attrs = patch.attrs
    if isinstance(required_attrs, Mapping):
        sub = {i: patch_attrs[i] for i in required_attrs}
        if sub != dict(required_attrs):
            msg = f"Patch's attrs {sub} are not required attrs: {required_attrs}"
            raise PatchAttributeError(msg)
    else:
        missing = set(required_attrs) - set(dict(patch.attrs))
        if missing:
            msg = f"Patch is missing the following attributes: {missing}"
            raise PatchAttributeError(msg)
    return patch


def patch_function(
    required_dims: Optional[Union[Tuple[str, ...], Callable]] = None,
    required_attrs: attr_type = None,
    history: Literal["full", "method_name", None] = "full",
):
    """
    Decorator to mark a function as a patch method.

    Parameters
    ----------
    required_dims
        A tuple of dimensions which must be found in the Patch.
    required_attrs
        A dict of attributes which must be found in the Patch and whose
        values must be equal to those provided.
    history
        Specifies how to track history on Patch.
            Full - Records function name and str version of input arguments.
            method_name - Only records method name. Useful if args are long.
            None - Function call is not recorded in history attribute.

    Examples
    --------
    1. A patch method which requires dimensions (time, distance)
    >>> import dascore as dc
    >>> @dc.patch_function(required_dims=('time', 'distance'))
    ... def do_something(patch):
    ...     ...   # raises a PatchCoordsError if patch doesn't have time,
    ...     #  distance

    2. A patch method which requires an attribute 'data_type' == 'DAS'
    >>> import dascore as dc
    >>> @dc.patch_function(required_attrs={'data_type': 'DAS'})
    ... def do_another_thing(patch):
    ...     ...  # raise PatchAttributeError if patch doesn't have attribute
    ...     # called "data_type" or its values is not equal to "DAS".

    Notes
    -----
    The original function can still be accessed with the .func attribute.
    This may be useful for avoiding calling the patch_func machinery
    multiple times from within another patch function.
    """

    def _wrapper(func):
        @functools.wraps(func)
        def _func(patch, *args, **kwargs):
            check_patch_dims(patch, required_dims)
            check_patch_attrs(patch, required_attrs)
            hist_str = _get_history_str(patch, func, *args, _history=history, **kwargs)
            out: PatchType = func(patch, *args, **kwargs)
            # attach history string. Consider something a bit less hacky.
            if hist_str and hasattr(out, "attrs"):
                out = _shallow_copy(out)
                out._data_array.attrs["history"].append(hist_str)
            return out

        _func.func = func  # attach original function
        _func.__wrapped__ = func

        return _func

    if callable(required_dims):  # the decorator is used without parens
        return patch_function()(required_dims)

    return _wrapper


def copy_attrs(attrs) -> dict:
    """Make a copy of the attributes dict so it can be mutated."""
    out = dict(attrs)
    if "history" in out:
        out["history"] = list(out["history"])
    return out


class _AttrsCoordsMixer:
    """Class for handling complex interactions between attrs and coords."""

    # a dict of {key: List[func]} for special handling of set attrs.
    set_attr_funcs = collections.defaultdict(list)
    # a dict of {key: func} for special handling of missing attributes.
    missing_attr_funcs = {}
    _missing_attr_keys = frozenset()
    # a dict of {attr: (expected_type, func)}
    _expected_types = dict(
        d_time=(np.timedelta64, to_timedelta64),
        time_min=(np.datetime64, to_datetime64),
        time_max=(np.datetime64, to_datetime64),
    )

    def __init__(self, attrs, coords, dims):
        """Init with attrs and coords"""
        if not dims and coords:
            dims = list(coords)
        self.attrs = attrs if attrs is not None else {}
        self.coords = Coords(coords)
        self.dims = dims if dims is not None else ()
        self._original_attrs = attrs
        self._original_coords = coords
        # fill missing values with default or implicit values
        self._set_default_attrs()
        self._set_attr_types()
        self._condition_coords()
        # infer any attr values from coordinates
        self._update_attrs_from_coords(self._missing_attr_keys)

    def update_attrs(self, **kwargs):
        """
        Update a coordinate.

        Parameters
        ----------
        **kwargs
            The coordinates to update.
        """
        for key, value in kwargs.items():
            attr = self.copied_attrs
            # convert types to those expected, mainly for converting times
            # from strings or ints to np.datetime64 or np.timedelta64.
            if key in self._expected_types:
                expected_type, func = self._expected_types[key]
                if not isinstance(value, expected_type):
                    value = func(value)
            attr[key] = value
            # apply special processing for this key
            for match_key in self._get_matching_patterns(self.set_attr_funcs, key):
                match_dims = [x for x in self.dims if x in key]
                for func in self.set_attr_funcs[match_key]:
                    func(self, match_dims=match_dims)

    def _get_matching_patterns(self, match_list, key):
        """
        Return any patterns that match key. Match list contains unix-style match
        strings.
        """
        for match_str in list(match_list):
            if re.match(translate(match_str), key):
                yield match_str

    @append_func(set_attr_funcs["*_min"])
    def _update_max_from_min(self, match_dims=()):
        """Update end coord based on new start coord."""
        if len(match_dims):
            assert len(match_dims) == 1
            dim = match_dims[0]
            td = self._original_attrs[f"{dim}_max"] - self._original_attrs[f"{dim}_min"]
            if not pd.isnull(td):
                attrs = self.copied_attrs
                attrs[f"{dim}_max"] = self.attrs[f"{dim}_min"] + td

    @append_func(set_attr_funcs["*_min"])
    def _update_coords_min(self, match_dims=()):
        """Update coordinate for new min"""
        if len(match_dims):
            assert len(match_dims) == 1
            dim = match_dims[0]
            coord = self.coords.get(dim)
            coord_min = self.attrs[f"{dim}_min"]
            if coord is not None and np.min(coord) != coord_min:
                td = coord - coord[0]
                self.coords = self.coords.update(**{dim: coord_min + td})

    @append_func(set_attr_funcs["*_max"])
    def _update_min_from_max(self, match_dims=()):
        """Update start based on new end."""
        if len(match_dims):
            assert len(match_dims) == 1
            dim = match_dims[0]
            td = self._original_attrs[f"{dim}_max"] - self._original_attrs[f"{dim}_min"]
            if not pd.isnull(td):
                attrs = self.copied_attrs
                attrs["time_min"] = self.attrs["time_max"] - td

    @append_func(set_attr_funcs["*_max"])
    def _update_coords_from_max(self, match_dims=()):
        """Update coordinate for new start of dimension."""
        if len(match_dims):
            assert len(match_dims) == 1
            dim = match_dims[0]
            coord = self.coords.get(dim, None)
            time_max = self.attrs[f"{dim}_max"]
            if coord is not None and np.max(coord) != time_max:
                td = coord - coord[-1]
                self.coords = self.coords.update(**{dim: time_max + td})

    @append_func(set_attr_funcs["d_*"])
    def _update_max_from_d(self, match_dims=()):
        """Update the ending attribute from updating dimension sampling."""
        if len(match_dims):
            assert len(match_dims) == 1
            dim = match_dims[0]
            coord = self.coords[dim]
            attrs = self.copied_attrs
            start = attrs.get(f"{dim}_min", None)
            if start:
                new_end = start + (len(coord) - 1) * attrs[f"d_{dim}"]
                self.copied_attrs[f"{dim}_max"] = new_end

    @append_func(set_attr_funcs["d_*"])
    def _update_coords_from_d(self, match_dims=()):
        """Update coordinates for new d_dim."""
        if len(match_dims):
            assert len(match_dims) == 1
            dim = match_dims[0]
            coord = self.coords.get(dim, None)
            d_dim = self.copied_attrs[f"d_{dim}"]
            start = self.copied_attrs.get(f"{dim}_min", None)
            if coord is not None and start is not None:
                new_coord = start + np.arange(len(coord)) * d_dim
                self.coords = self.coords.update(**{dim: new_coord})

    @functools.cached_property
    def copied_attrs(self):
        """Make a copy of the attrs before mutating."""
        self.attrs = copy_attrs(self.attrs)
        return self.attrs

    @functools.cached_property
    def copied_coords(self):
        """Make a copy of the coords before mutating."""
        # TODO I don't like having this deep copy but shallow still caused
        return copy.deepcopy(self.coords)

    def __call__(self, *args, **kwargs):
        return PatchAttrs(**self.attrs), self.coords

    def _set_default_attrs(self):
        """Get the attribute dict, add required keys if not yet defined."""
        # add default values if they are not in out or attrs yet
        defaults = PatchAttrs.get_defaults()
        defaults.update(self.attrs)
        self.attrs = defaults

    # --- Coordinate bits

    def _update_attrs_from_coords(self, fill_keys=None):
        """Fill in missing attributes that can be inferred."""
        fill_keys = fill_keys if fill_keys else set(self.attrs)
        # iterate each dimension, fill in missing values with data from coords
        for dim in self.dims:
            array = self.coords[dim]
            # make sure its not an xarray thing
            array = getattr(array, "values", array)
            if f"{dim}_min" in fill_keys:
                minval = np.min(array) if len(array) else np.NaN
                self.copied_attrs[f"{dim}_min"] = minval
            if f"{dim}_max" in fill_keys:
                maxval = np.max(array) if len(array) else np.NaN
                self.copied_attrs[f"{dim}_max"] = maxval
        # add dims column
        if not len(self.attrs["dims"]):
            self.copied_attrs["dims"] = ",".join(self.dims)

    def update_coords(self, **kwargs):
        """Update the coordinates based on kwarg inputs."""
        self.coords = self.coords.update(**kwargs)
        self._condition_coords()
        self._update_attrs_from_coords()

    def _condition_coords(self):
        """
        Condition the coordinates before using them.

        This is mainly to enforce common time conventions
        """
        coords = self.coords
        if coords is None or (time := coords.get("time")) is None:
            return
        start_time = self.attrs["time_min"]
        if pd.isnull(start_time):
            start_time = to_datetime64(0)
        # Convert non-datetime into time deltas
        if not np.issubdtype(time.dtype, np.datetime64):
            td = to_timedelta64(time)
            time = start_time + td
        self.coords = self.coords.update(time=time)

    def _set_attr_types(self):
        """Make sure time attrs are expected types"""
        for key, (expected_type, func) in self._expected_types.items():
            val = self.attrs[key]
            if not isinstance(val, expected_type):
                self.copied_attrs[key] = func(val)


@compose_docstring(fields=PatchAttrs.__annotations__)
def patches_to_df(
    patches: Union[Sequence[PatchType], SpoolType, pd.DataFrame]
) -> pd.DataFrame:
    """
    Return a dataframe

    Parameters
    ----------
    patches
        A sequence of :class:`dascore.Patch`

    Returns
    -------
    A dataframe with the following fields:
        {fields}
    plus a field called 'patch' which contains a reference to each patch.
    """

    if isinstance(patches, dc.BaseSpool):
        df = patches._df
    # Handle spool case
    elif hasattr(patches, "get_contents"):
        return patches.get_contents()
    elif isinstance(patches, pd.DataFrame):
        return patches
    else:
        df = pd.DataFrame([dict(x) for x in scan_patches(patches)])
        if df.empty:  # create empty df with appropriate columns
            cols = list(PatchAttrs().dict())
            df = pd.DataFrame(columns=cols).assign(patch=None, history=None)
        else:  # else populate with patches and concat history
            history = df["history"].apply(lambda x: ",".join(x))
            df = df.assign(patch=patches, history=history)
    # Ensure history is in df
    if "history" not in df.columns:
        df = df.assign(history="")
    if "patch" not in df.columns:
        df["patch"] = [x for x in patches]
    return df


def merge_patches(
    patches: Union[Sequence[PatchType], pd.DataFrame, SpoolType],
    dim: str = "time",
    check_history: bool = True,
    tolerance: float = 1.5,
) -> Sequence[PatchType]:
    """
    Merge all compatible patches in spool or patch list together.

    Parameters
    ----------
    patches
        A sequence of patches to merge (if compatible)
    dim
        The dimension along which to merge
    check_history
        If True, only merge patches with common history. This will, for
        example, prevent merging filtered and unfiltered data together.
    tolerance
        The upper limit of a gap to tolerate in terms of the sampling
        along the desired dimension. E.G., the default value means any patches
        with gaps <= 1.5 * dt will be merged.
    """
    msg = (
        "merge_patches is deprecated. Use spool.chunk instead."
        "For example, to merge a list of patches you can use:"
        "dascore.spool(patch_list).chunk(time=None) to merge on the time "
        "dimension"
    )
    warnings.warn(msg, DeprecationWarning)
    return _merge_patches(patches, dim, check_history, tolerance)


def _merge_patches(
    patches: Union[Sequence[PatchType], pd.DataFrame, SpoolType],
    dim: str = "time",
    check_history: bool = True,
    tolerance: float = 1.5,
) -> Sequence[PatchType]:
    """
    Merge all compatible patches in spool or patch list together.

    Parameters
    ----------
    patches
        A sequence of patches to merge (if compatible)
    dim
        The dimension along which to merge
    check_history
        If True, only merge patches with common history. This will, for
        example, prevent merging filtered and unfiltered data together.
    tolerance
        The upper limit of a gap to tolerate in terms of the sampling
        along the desired dimension. E.G., the default value means any patches
        with gaps <= 1.5 * dt will be merged.
    """

    def _get_sorted_df_sort_group_names(patches):
        """Return the sorted dataframe."""
        group_names = list(PATCH_MERGE_ATTRS) + [d_name]
        sort_names = group_names + [min_name, max_name]
        if check_history:
            sort_names += ["history"]
            patches = patches.assign(history=lambda x: x["history"].apply(str))
        return patches.sort_values(sort_names), sort_names, group_names

    def _merge_compatible_patches(patch_df):
        """perform merging after patch compatibility has been confirmed."""
        has_overlap = patch_df["_dist_to_previous"] <= to_timedelta64(0)
        overlap_start = patch_df[min_name] - patch_df["_dist_to_previous"]
        # get data arrays
        dars = [x._data_array for x in patch_df["patch"]]
        dars = [
            _trim_or_fill(x, start) if needs_action else x
            for x, start, needs_action in zip(dars, overlap_start, has_overlap)
        ]
        dar = xr.concat(dars, dim=dim)
        dar.attrs[min_name] = np.NaN
        dar.attrs[max_name] = np.NaN
        dims = tuple(dar.dims)
        return dc.Patch(dar.data, coords=dar.coords, attrs=dar.attrs, dims=dims)

    def _trim_or_fill(dar, new_start):
        """Trim or fill data array."""
        return dar.loc[{dim: dar.coords[dim] > new_start}]

    # get a dataframe
    if not isinstance(patches, pd.DataFrame):
        patches = patches_to_df(patches)
    assert dim in {"time", "distance"}, "merge must be on time/distance for now"
    out = []  # list of merged patches
    min_name, max_name, d_name = f"{dim}_min", f"{dim}_max", f"d_{dim}"
    # get sorted dataframe and group/sort column names
    df, sorted_names, group_names = _get_sorted_df_sort_group_names(patches)
    # get a boolean if each row is compatible with previous for merging
    gn = ~(df[group_names] == df.shift()[group_names]).all(axis=1)
    group_numbers = gn.astype(bool).cumsum()
    for _, sub_df in df.groupby(group_numbers):
        # get boolean indicating if patch overlaps with previous
        start, end = sub_df[min_name], sub_df[max_name]
        merge_dist = sub_df[d_name] * tolerance
        # get the dist between each patch
        dist_to_previous = start - end.shift()
        no_merge = ~(dist_to_previous <= merge_dist)
        sub_df["_dist_to_previous"] = dist_to_previous
        # determine if each patch should be merged with the previous one
        for _, merge_patch_df in sub_df.groupby(no_merge.astype(int).cumsum()):
            out.append(_merge_compatible_patches(merge_patch_df))
    return out


@compose_docstring(fields=format_dtypes(PatchFileSummary.__annotations__))
def scan_patches(
    patches: Union[PatchType, Sequence[PatchType]]
) -> list[PatchFileSummary]:
    """
    Scan a sequence of patches and return a list of summaries.

    The summary dicts have the following fields:
        {fields}

    Parameters
    ----------
    patches
        A single patch or a sequence of patches.
    """
    if isinstance(patches, dc.Patch):
        patches = [patches]  # make sure we have an iterable
    out = [PatchFileSummary(**dict(pa.attrs)) for pa in patches]
    return out


def get_start_stop_step(patch: PatchType, dim):
    """
    Convenience method for getting start, stop, step for a given dimension.
    """
    assert dim in patch.dims, f"{dim} is not in Patch dimensions of {patch.dims}"
    start = patch.attrs[f"{dim}_min"]
    end = patch.attrs[f"{dim}_max"]
    step = patch.attrs[f"d_{dim}"]
    return start, end, step


def get_default_patch_name(patch):
    """Generates the name of the node."""

    def _format_datetime64(dt):
        """Format the datetime string in a sensible way."""
        out = str(np.datetime64(dt).astype("datetime64[ns]"))
        return out.replace(":", "_").replace("-", "_").replace(".", "_")

    attrs = patch.attrs
    start = _format_datetime64(attrs.get("time_min", ""))
    end = _format_datetime64(attrs.get("time_max", ""))
    net = attrs.get("network", "")
    sta = attrs.get("station", "")
    tag = attrs.get("tag", "")
    return f"DAS__{net}__{sta}__{tag}__{start}__{end}"


def get_dim_value_from_kwargs(patch, kwargs):
    """
    Assert that kwargs contain one value and it is a dimension of patch.

    Several patch functions allow passing values via kwargs which are dimension
    specific. This function allows for some sane validation of such functions.

    Return the name of the dimension, its axis position, and its value.
    """
    dims = patch.dims
    overlap = set(dims) & set(kwargs)
    if len(kwargs) != 1 or not overlap:
        msg = (
            "You must use exactly one dimension name in kwargs. "
            f"You passed the following kwargs: {kwargs} to a patch with "
            f"dimensions {patch.dims}"
        )
        raise PatchDimError(msg)
    dim = list(overlap)[0]
    axis = dims.index(dim)
    return dim, axis, kwargs[dim]



"""
Utilities for TDMS format.
"""
import datetime
import mmap
import os
import struct

import numpy as np

from dascore.core.schema import PatchAttrs
from dascore.utils.misc import get_slice
from dascore.utils.time import to_datetime64

DEFAULT_ATTRS = tuple(PatchAttrs.__fields__)


def type_not_supported(vargin):
    """Function raises a NotImplementedException."""
    raise NotImplementedError("Reading of this tdsDataType is not implemented")


def parse_time_stamp(fractions, seconds):
    """
    Convert time TDMS time representation to datetime
    fractions   -- fractional seconds (2^-64)
    seconds     -- The number of seconds since 1/1/1904
    @rtype : datetime.datetime
    """
    if fractions is not None and seconds is not None and fractions + seconds > 0:
        return datetime.timedelta(
            0, fractions * 2**-64 + seconds
        ) + datetime.datetime(1904, 1, 1)
    else:
        return None


def _get_version_str(tdms_file, LEAD_IN_LENGTH=28) -> str:
    """Return True if we have the right file type"""
    lead_in = tdms_file.read(LEAD_IN_LENGTH)
    # lead_in is 28 bytes:
    # [string of length 4][int32][int32][int64][int64]
    fields = struct.unpack("<4siiQQ", lead_in)

    # TODO: validate file
    if fields[0].decode() in "TDSm":
        version_str = str(fields[2])
        return version_str
    else:
        return False


def _get_distance_array(tdms_file=None, attrs=None):
    """Get the distance (along fiber) array."""
    # Note: At least for the test file, sensing_range_start, sensing_range_stop,
    # nx, and dx are not consistent so I just used this method. We need to
    # look more into this.
    if attrs is None:
        attrs, _ = _get_all_attrs(tdms_file, LEAD_IN_LENGTH=28)
    zero_offset = attrs["distance_min"]
    channel_spacing = attrs["d_distance"] * attrs["Fibre Length Multiplier"]
    max_dist = attrs["distance_max"]
    # Note: I found files where the end was slightly less than one channel spacing
    # as a result the distance array was one channels short. Adding 10% of the
    # channels spacing to distance_max fixes the issue without causing problems
    # for files that don't have this rounding error.
    dist = np.arange(zero_offset, max_dist + channel_spacing / 10, channel_spacing)
    return dist


def _get_time_array(tdms_file=None, attrs=None):
    """Get the time array for the file."""
    if attrs is None:
        attrs, _ = _get_all_attrs(tdms_file, LEAD_IN_LENGTH=28)
    time_min = attrs["time_min"]
    time_max = attrs["time_max"]
    time_array = np.arange(
        time_min, time_max, np.timedelta64(int(1000 * attrs["d_time"]), "ms")
    )
    return time_array


def _get_default_attrs(tdms_file, get_all_attrs=None):
    """
    Return the required/default attributes which can be fetched from attributes.

    """
    if get_all_attrs is None:
        all_attrs, _ = _get_all_attrs(tdms_file)
    else:
        all_attrs = get_all_attrs
    out = {
        default_attr: all_attrs[default_attr]
        for default_attr in DEFAULT_ATTRS
        if default_attr in all_attrs
    }
    return out


def _get_all_attrs(tdms_file, LEAD_IN_LENGTH=28):
    """
    Return all the attributes which can be fetched from attributes.

    """
    # Enum mapping TDM data types to description string, numpy type where exists
    # See Ref[2] for enum values
    TDS_DATA_TYPE = dict(
        {
            0x00: "void",  # tdsTypeVoid
            0x01: "int8",  # tdsTypeI8
            0x02: "int16",  # tdsTypeI16
            0x03: "int32",  # tdsTypeI32
            0x04: "int64",  # tdsTypeI64
            0x05: "uint8",  # tdsTypeU8
            0x06: "uint16",  # tdsTypeU16
            0x07: "uint32",  # tdsTypeU32
            0x08: "uint64",  # tdsTypeU64
            0x09: "float32",  # tdsTypeSingleFloat
            0x0A: "float64",  # tdsTypeDoubleFloat
            0x0B: "float128",  # tdsTypeExtendedFloat
            0x19: "singleFloatWithUnit",  # tdsTypeSingleFloatWithUnit
            0x1A: "doubleFloatWithUnit",  # tdsTypeDoubleFloatWithUnit
            0x1B: "extendedFloatWithUnit",  # tdsTypeExtendedFloatWithUnit
            0x20: "str",  # tdsTypeString
            0x21: "bool",  # tdsTypeBoolean
            0x44: "datetime",  # tdsTypeTimeStamp
            0xFFFFFFFF: "raw",  # tdsTypeDAQmxRawData
        }
    )

    # Function mapping for reading TDMS data types
    TDS_READ_VAL = dict(
        {
            "void": lambda f: None,  # tdsTypeVoid
            "int8": lambda f: struct.unpack("<b", f.read(1))[0],
            "int16": lambda f: struct.unpack("<h", f.read(2))[0],
            "int32": lambda f: struct.unpack("<i", f.read(4))[0],
            "int64": lambda f: struct.unpack("<q", f.read(8))[0],
            "uint8": lambda f: struct.unpack("<B", f.read(1))[0],
            "uint16": lambda f: struct.unpack("<H", f.read(2))[0],
            "uint32": lambda f: struct.unpack("<I", f.read(4))[0],
            "uint64": lambda f: struct.unpack("<Q", f.read(8))[0],
            "float32": lambda f: struct.unpack("<f", f.read(4))[0],
            "float64": lambda f: struct.unpack("<d", f.read(8))[0],
            "float128": type_not_supported,
            "singleFloatWithUnit": type_not_supported,
            "doubleFloatWithUnit": type_not_supported,
            "extendedFloatWithUnit": type_not_supported,
            "str": lambda f: f.read(struct.unpack("<i", f.read(4))[0]),
            "bool": lambda f: struct.unpack("<?", f.read(1))[0],
            "datetime": lambda f: parse_time_stamp(
                struct.unpack("<Q", f.read(8))[0], struct.unpack("<q", f.read(8))[0]
            ),
            "raw": type_not_supported,
        }
    )

    DECIMATE_MASK = 0b00100000
    FILEINFO_NAMES = (
        "file_tag",
        "toc",
        "version",
        "next_segment_offset",
        "raw_data_offset",
    )

    # read leadin infomation into fileinfo
    lead_in = tdms_file.read(LEAD_IN_LENGTH)
    # lead_in is 28 bytes:
    fields = struct.unpack("<4siiQQ", lead_in)

    # Keep track of information about file in fileinfo
    fileinfo = dict(zip(FILEINFO_NAMES, fields))
    fileinfo["decimated"] = not bool(fileinfo["toc"] & DECIMATE_MASK)
    # Make offsets relative to beginning of file:
    fileinfo["next_segment_offset"] += LEAD_IN_LENGTH
    fileinfo["raw_data_offset"] += LEAD_IN_LENGTH
    fileinfo["file_size"] = os.path.getsize(tdms_file.name)

    # Make sure next segment does not go beyond file capacity
    if fileinfo["next_segment_offset"] > fileinfo["file_size"]:
        fileinfo["next_segment_offset"] = fileinfo["file_size"]

    def _read_attr(tdms_file):
        """
        Read a single property from the TDMS file.
        Return the name, type and value of the property as a list.
        """

        # Read length of object path:
        var = struct.unpack("<i", tdms_file.read(4))[0]
        # Read property name and type:
        name, data_type = struct.unpack("<{0}si".format(var), tdms_file.read(var + 4))
        # Lookup function to read and parse property value based on type:
        value = TDS_READ_VAL[TDS_DATA_TYPE[data_type]](tdms_file)
        name = name.decode()
        if data_type == 32:
            value = value.decode()

        return name, value  # data_type, value

    # navigate pointer to immediately after lead in data
    tdms_file.seek(LEAD_IN_LENGTH, 0)
    # Read number of channels
    n_channels = struct.unpack("i", tdms_file.read(4))[0] - 2
    fileinfo["n_channels"] = n_channels
    # Read length of object path:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    # skip over object path and raw data index:
    tdms_file.seek(var + 4, 1)
    # Read number of properties in this group:
    var = struct.unpack("<i", tdms_file.read(4))[0]

    # loop through and read each property
    out = [_read_attr(tdms_file) for _ in range(var)]
    # Returns a pandas dataframe that we convert to dictionary
    out = dict(out)
    # Add other attributes not yet included
    out["n_channels"] = n_channels
    out["data_type"] = "strain_rate"
    out["data_units"] = ""
    out["dims"] = "time, distance"
    out["d_time"] = 1 / out["SamplingFrequency[Hz]"]
    out["time_min"] = to_datetime64(str(out["GPSTimeStamp"]))
    # Rename some attributes to preferred names
    _root_attrs = {
        "SpatialResolution[m]": "d_distance",
        "SystemInfomation.Devices0.SerialNum": "instrument_id",
        "Start Distance (m)": "distance_min",
        "Stop Distance (m)": "distance_max",
    }
    for treble_name, out_name in _root_attrs.items():
        out[out_name] = out[treble_name]
        del out[treble_name]

    fileinfo["end_of_properties_offset"] = tdms_file.tell()

    tdms_file.seek(fileinfo["end_of_properties_offset"], 0)

    # skip over Group Information:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    tdms_file.seek(var + 8, 1)

    # skip over first channel path and length of index information:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    tdms_file.seek(var + 4, 1)

    fileinfo["data_type"] = TDS_DATA_TYPE.get(struct.unpack("<i", tdms_file.read(4))[0])

    if fileinfo["data_type"] not in ("int16", "float32"):
        raise Exception("Unsupported TDMS data type: " + fileinfo["data_type"])
    # get number of samples by dividing amount of unread data by the
    # size of data per channel
    numofsamples = (
        (fileinfo["file_size"] - fileinfo["raw_data_offset"])
        / n_channels
        / np.dtype(fileinfo["data_type"]).itemsize
    )
    out["time_max"] = out["time_min"] + np.timedelta64(
        int(1000 * numofsamples * out["d_time"]), "ms"
    )

    return out, fileinfo


def _get_fileinfo(tdms_file, LEAD_IN_LENGTH=28):
    """
    get info about file not included in the attributes

    """
    attrs, fileinfo = _get_all_attrs(tdms_file)

    # Read Dimension of the raw data array (has to be 1):
    _ = struct.unpack("<i", tdms_file.read(4))[0]

    fileinfo["chunk_size"] = struct.unpack("<i", tdms_file.read(4))[0]

    return fileinfo, attrs


def _get_data_node(tdms_file, LEAD_IN_LENGTH=28):
    """Get all the data saved in the current file"""

    def get_segmentdata(fileinfo, nch, dmap, nso, rdo):

        # seg1_length: length of recording indicated as raw_data in metadata for
        # each channel in bytes
        seg_length = int((nso - rdo) / nch / np.dtype(fileinfo["data_type"]).itemsize)
        channel_length = seg_length

        if fileinfo["decimated"]:
            # number of completely full chunks
            n_complete_blk = int(seg_length / fileinfo["chunk_size"])
            ax_ord = "C"
        else:
            n_complete_blk = 0
            ax_ord = "F"
        # use data from mapped file to fill variable raw_data
        raw_data = np.ndarray(
            (n_complete_blk, nch, fileinfo["chunk_size"]),
            dtype=fileinfo["data_type"],
            buffer=dmap,
            offset=rdo,
        )
        # Rotate the axes to [chunk_size, nblk, nch]
        raw_data = np.rollaxis(raw_data, 2)
        data_node = np.reshape(raw_data, (n_complete_blk * fileinfo["chunk_size"], nch))
        if n_complete_blk != seg_length / fileinfo["chunk_size"]:
            # If the last chunk isn't full there is some data left
            additional_samples = int(
                seg_length - n_complete_blk * fileinfo["chunk_size"]
            )
            additional_samples_offset = (
                rdo
                + n_complete_blk
                * nch
                * fileinfo["chunk_size"]
                * np.dtype(fileinfo["data_type"]).itemsize
            )
            raw_last_chunk = np.ndarray(
                (nch, additional_samples),
                dtype=fileinfo["data_type"],
                buffer=dmap,
                offset=additional_samples_offset,
                order=ax_ord,
            )
            # Rotate the axes to [samples, nch]
            raw_last_chunk = np.rollaxis(raw_last_chunk, 1)
            data_node = np.append(data_node, raw_last_chunk, axis=0)
            return data_node, channel_length

    fileinfo, attrs = _get_fileinfo(tdms_file)

    # map file contents to a variable dmap
    dmap = mmap.mmap(tdms_file.fileno(), 0, access=mmap.ACCESS_READ)
    # rdo: the start of the data in the file
    rdo = int(fileinfo["raw_data_offset"])
    # nch: number of channels
    nch = int(fileinfo["n_channels"])

    # nso: the beginning of the segment that comes next after the data
    nso = fileinfo["next_segment_offset"]
    # seg1_length: length of recording indicated as raw_data in metadata for
    # each channel in bytes

    flag = 0
    while flag == 0:
        cdata_node, cchannel_length = get_segmentdata(fileinfo, nch, dmap, nso, rdo)
        if fileinfo["file_size"] == nso:
            flag = 1
        else:
            tdms_file.seek(nso + 12, 0)
            (next_seg_nso, next_seg_rdo) = struct.unpack("<qq", tdms_file.read(2 * 8))
            rdo = min(fileinfo["file_size"], nso + LEAD_IN_LENGTH + next_seg_rdo)
            nso = min(fileinfo["file_size"], nso + LEAD_IN_LENGTH + next_seg_nso)
        try:
            data_node = np.append(data_node, cdata_node, axis=1)
            channel_length += cchannel_length
        except NameError:
            data_node = cdata_node
            channel_length = cchannel_length

    return data_node, channel_length, attrs


def _get_data(time, distance, time_array, dist_array, data_node):
    """
    Get the data array. Slice based on input and check for 0 blocks. Also
    return sliced coordinates.
    """
    # need to handle empty data blocks. This happens when data is stopped
    # recording before the pre-allocated file is filled.
    if time_array[-1] < time_array[0]:
        time = (time[0], time_array.max())
    tslice = get_slice(time_array, time)
    dslice = get_slice(dist_array, distance)
    return data_node[tslice, dslice], time_array[tslice], dist_array[dslice]



"""
Compatibility module for DASCore.

All components/functions that may be exchanged for other numpy/scipy
compatible libraries should go in this model.
"""

from numpy import array, floor, interp  # NOQA
from scipy.interpolate import interp1d  # NOQA
from scipy.ndimage import zoom  # NOQA
from scipy.signal import decimate, resample, resample_poly  # NOQA



"""
DASDAE format utilities
"""
import numpy as np

import dascore as dc
from dascore.core.schema import PatchFileSummary
from dascore.utils.hdf5 import open_hdf5_file
from dascore.utils.patch import get_default_patch_name
from dascore.utils.time import to_number, to_timedelta64

# --- Functions for writing DASDAE format.


def _write_meta(hfile, file_version):
    """Write metadata to hdf5 file."""
    attrs = hfile.root._v_attrs
    attrs["__format__"] = "DASDAE"
    attrs["__DASDAE_version__"] = file_version
    attrs["__dascore__version__"] = dc.__version__


def _save_attrs_and_dims(patch, patch_group):
    """Save the attributes."""
    # copy attrs to group attrs
    # TODO will need to test if objects are serializable
    for i, v in patch.attrs.items():
        patch_group._v_attrs[f"_attrs_{i}"] = v
    patch_group._v_attrs["_dims"] = ",".join(patch.dims)


def _save_array(data, name, group, h5):
    """Save an array to a group, handle datetime flubbery."""
    # handle datetime conversions
    is_dt = np.issubdtype(data.dtype, np.datetime64)
    is_td = np.issubdtype(data.dtype, np.timedelta64)
    if is_dt or is_td:
        data = to_number(data)
    out = h5.create_array(
        group,
        name,
        data,
    )
    out._v_attrs["is_datetime64"] = is_dt
    out._v_attrs["is_timedelta64"] = is_td


def _save_coords(patch, patch_group, h5):
    """Save coordinates"""
    for name, coord in patch._data_array.coords.items():
        # First save coordinate arrays
        data = coord.values
        save_name = f"_coord_{name}"
        _save_array(data, save_name, patch_group, h5)
        # then save dimensions of coordinates
        save_name = f"_cdims_{name}"
        patch_group._v_attrs[save_name] = ",".join(coord.dims)


def _save_patch(patch, wave_group, h5):
    """Save the patch to disk."""
    name = get_default_patch_name(patch)
    patch_group = h5.create_group(wave_group, name)
    _save_attrs_and_dims(patch, patch_group)
    _save_coords(patch, patch_group, h5)
    # add data
    if patch.data.shape:
        h5.create_array(patch_group, "data", patch.data)


# --- Functions for reading


def _get_attrs(patch_group):
    """Get the saved attributes form the group attrs."""
    out = {}
    attrs = [x for x in patch_group._v_attrs._f_list() if x.startswith("_attrs_")]
    for attr_name in attrs:
        key = attr_name.replace("_attrs_", "")
        val = patch_group._v_attrs[attr_name]
        # need to unpack one value arrays
        if isinstance(val, np.ndarray) and not val.shape:
            val = np.array([val])[0]
        out[key] = val
        #
    return out


def _read_array(table_array):
    """Read an array into numpy."""
    data = table_array[:]
    if table_array._v_attrs["is_datetime64"]:
        data = data.astype("datetime64[ns]")
    if table_array._v_attrs["is_timedelta64"]:
        data = to_timedelta64(data)
    return data


def _get_coords(patch_group):
    """Get the coordinates from a patch group."""
    out = {}
    for coord in [x for x in patch_group if x.name.startswith("_coord_")]:
        name = coord.name.replace("_coord_", "")
        out[name] = _read_array(coord)

    attrs = [x for x in patch_group._v_attrs._f_list() if x.startswith("_cdims")]
    for coord_name in attrs:
        name = coord_name.replace("_cdims_", "")
        value = patch_group._v_attrs[coord_name]
        assert name in out, "Should already have loaded coordinate array"
        # add dimensions to coordinates that have them.
        out[name] = (tuple(value.split(".")), out[name])
    return out


def _get_dims(patch_group):
    """Get the dims tuple from the patch group."""
    dims = patch_group._v_attrs["_dims"]
    if not dims:
        out = ()
    else:
        out = tuple(dims.split(","))
    return out


def _read_patch(patch_group, **kwargs):
    """Read a patch group, return Patch."""
    attrs = _get_attrs(patch_group)
    dims = _get_dims(patch_group)
    coords = _get_coords(patch_group)
    try:
        data = patch_group["data"][:]
    except (IndexError, KeyError):
        data = np.array(None)
    return dc.Patch(data=data, coords=coords, dims=dims, attrs=attrs)


def _get_contents_from_patch_groups(path, file_version, file_format="DASDAE"):
    """Get the contents from each patch group."""
    out = []
    with open_hdf5_file(path) as h5:
        for group in h5.iter_nodes("/waveforms"):
            contents = _get_patch_content_from_group(group)
            # populate file info
            contents["file_version"] = file_version
            contents["file_format"] = file_format
            contents["path"] = h5.filename
            out.append(PatchFileSummary(**contents))
    return out


def _get_patch_content_from_group(group):
    """Get patch content from a single node."""
    attrs = group._v_attrs
    out = {i.replace("_attrs_", ""): getattr(attrs, i) for i in attrs._f_list()}
    # rename dims
    out["dims"] = out.pop("_dims")
    return out



"""
Module for static visualizations and figure generation.

For interactive visualizations see :module:`dascore.workbench`
"""
from dascore.utils.misc import MethodNameSpace

from .spectrogram import spectrogram
from .waterfall import waterfall
from .wiggle import wiggle


class VizPatchNameSpace(MethodNameSpace):
    """A class for storing visualization namespace."""

    waterfall = waterfall
    spectrogram = spectrogram
    wiggle = wiggle



"""
Tests for waterfall plots.
"""
import matplotlib.pyplot as plt
import pytest

import dascore as dc


@pytest.fixture(scope="session")
def patch_random_start(event_patch_1):
    """Get a patch with a random, odd, starttime."""
    random_starttime = dc.to_datetime64("2020-01-02T02:12:11.02232")
    attrs = dict(event_patch_1.attrs)
    coords = {i: v for i, v in event_patch_1.coords.items()}
    time = coords["time"] - coords["time"].min()
    coords["time"] = time + random_starttime
    attrs["time_min"] = coords["time"].min()
    attrs["time_max"] = coords["time"].max()
    patch = event_patch_1.new(attrs=attrs, coords=coords)
    return patch


class TestWaterfall:
    """Tests for waterfall plot"""

    def test_returns_axes(self, random_patch):
        """Call waterfall plot, return."""
        # modify patch to include line at start
        data = random_patch.data
        data[:100, :100] = 2.0  # create an origin block for testing axis line up
        data[:100, -100:] = -2.0  #
        out = random_patch.new(data=data)
        ax = out.viz.waterfall()

        # check labels
        assert random_patch.dims[0].capitalize() in ax.get_ylabel()
        assert random_patch.dims[1].capitalize() in ax.get_xlabel()
        assert isinstance(ax, plt.Axes)

    def test_colorbar_scale(self, random_patch):
        """Tests for the scaling parameter."""
        ax_scalar = random_patch.viz.waterfall(scale=0.2)
        assert ax_scalar is not None
        seq_scalar = random_patch.viz.waterfall(scale=[0.1, 0.3])
        assert seq_scalar is not None

    def test_colorbar_absolute_scale(self, random_patch):
        """Tests for absolute scaling of colorbar."""
        patch = random_patch.new(data=random_patch.data * 100 - 50)
        ax1 = patch.viz.waterfall(scale_type="absolute", scale=(-50, 50))
        assert ax1 is not None
        ax2 = patch.viz.waterfall(scale_type="absolute", scale=10)
        assert ax2 is not None

    def test_doc_intro_example(self, event_patch_1):
        """Simple test to ensure the doc examples can be run"""
        patch = event_patch_1.pass_filter(time=(None, 300))
        _ = patch.viz.waterfall(scale=0.04)
        _ = patch.transpose("distance", "time").viz.waterfall(scale=0.04)



"""Pydantic schemas used by DASCore."""
from pathlib import Path
from typing import Literal, Sequence, Union

import numpy as np
from pydantic import BaseModel, Field

from dascore.constants import (
    VALID_DATA_CATEGORIES,
    VALID_DATA_TYPES,
    basic_summary_attrs,
    max_lens,
)
from dascore.utils.docs import compose_docstring
from dascore.utils.time import to_datetime64, to_timedelta64


class SimpleValidator:
    """
    A custom class for getting simple validation behavior in pydantic.

    Subclass, then define function to be used as validator. func
    """

    @classmethod
    def func(cls, value):
        """A method to overwrite with custom validation."""
        return value

    @classmethod
    def __get_validators__(cls):
        """Hook used by pydantic."""
        yield cls.validate

    @classmethod
    def validate(cls, validator):
        """Simply call func."""
        return cls.func(validator)


class DateTime64(np.datetime64, SimpleValidator):
    """DateTime64 validator"""

    func = to_datetime64


class TimeDelta64(np.timedelta64, SimpleValidator):
    """TimeDelta64 validator"""

    func = to_timedelta64


@compose_docstring(basic_params=basic_summary_attrs)
class PatchAttrs(BaseModel):
    """
    The expected attributes for a Patch.

    Parameter
    ---------
    {basic_params}

    Notes
    -----
    These attributes go into the HDF5 index used by dascore. Therefore,
    when they are changed the index version needs to be incremented so
    previous indices are invalidated.

    See also [PatchAttrs](`dascore.core.schema.PatchAttrs`).
    """

    data_type: Literal[VALID_DATA_TYPES] = ""
    data_category: Literal[VALID_DATA_CATEGORIES] = ""
    data_units: str = ""
    time_min: DateTime64 = np.datetime64("NaT")
    time_max: DateTime64 = np.datetime64("NaT")
    d_time: TimeDelta64 = np.timedelta64("NaT")
    time_units: str = ""
    distance_min: float = np.NaN
    distance_max: float = np.NaN
    d_distance: float = np.NaN
    distance_units: str = ""
    instrument_id: str = Field("", max_length=max_lens["instrument_id"])
    cable_id: str = Field("", max_length=max_lens["cable_id"])
    dims: Union[tuple[str, ...], str] = tuple()
    tag: str = Field("", max_length=max_lens["tag"])
    station: str = Field("", max_length=max_lens["station"])
    network: str = Field("", max_length=max_lens["network"])
    history: Union[str, Sequence[str]] = Field(default_factory=list)

    # In order to maintain backward compatibility, these dunders make the
    # class also behave like a dict.

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __len__(self):
        return len(self.dict())

    def get(self, item, default=None):
        """dict-like get method."""
        try:
            return self[item]
        except (AttributeError, KeyError):
            return default

    def items(self):
        """Yield (attribute, values) just like dict.items()."""
        for item, value in self.dict().items():
            yield item, value

    @classmethod
    def get_defaults(cls):
        """return a dict of default values"""
        new = cls()
        return new.dict()

    class Config:
        """Configuration for Patch Summary"""

        title = "Patch Summary"
        extra = "allow"
        allow_mutation = False
        json_encoders = {
            np.datetime64: lambda x: str(x),
            np.timedelta64: lambda x: str(x),
        }


class PatchFileSummary(PatchAttrs):
    """
    The expected minimum attributes for a Patch/spool file.
    """

    # These attributes are excluded from the HDF index.
    _excluded_index = ("data_units", "time_units", "distance_units", "history")

    file_version: str = ""
    file_format: str = ""
    path: Union[str, Path] = ""

    @classmethod
    def get_index_columns(cls) -> tuple[str, ...]:
        """Return the column names which should be used for indexing."""
        fields = set(cls.__fields__) - set(cls._excluded_index)
        return tuple(sorted(fields))
